﻿/*
 * ##############################################
 * Calculation
 * Main part of the QuickCalculator
 * ##############################################
 * 
 * You may copy and distribute this code and the 
 * whole program as you want.
 * This code is not meant to be sold.
 * 
 * Written by Philipp Erler
 * Contact me for feedback, new version or other causes:
 * ph.erler@gmx.net
 * 
 * Version info:
 * Last change: 2013 | 02 | 03
 * .net-Framework: V. 4.0
 * 
 */

using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Globalization;

namespace CalculationE
{
    static public class CalcEngine
    {
        #region Header_Con_Destructor
        public enum angleModes                  { Radian, Degree };
	    static private angleModes angleMode;

        public enum DecimalSign                 { comma, point };
        static private DecimalSign currCult;
        static private CultureInfo[] cultures = { new CultureInfo("en-US"), new CultureInfo("de-DE") };

        static private string calcLog = "";
        static private bool writeLog = false;

        public enum OutputFormat                { Exponential, Decimal, Binary, Octal, Hexadecimal };
        static private int numRoundDecimals =   10;                             //only valid between 0 and 15
        static private OutputFormat outputFormat = OutputFormat.Decimal;
        private enum CtrlStructType             { IF, EndIf, ElseIf, Else, While, Break, Continue, Loop, None };
        const string IF = "if";
        const string ELSEIF = "elseif";
        const string ELSE = "else";
        const string ENDIF = "endif";
        const string WHILE = "while";
        const string BREAK = "break";
        const string CONTINUE = "continue";
        const string LOOP = "loop";

        //constructor
        static CalcEngine()
        {
            angleMode = angleModes.Radian;                                      //set anglemode to rad
            Default_Culture();
        }
        #endregion

        #region Calc_Log

        /// <summary>
        /// Set use log switch.
        /// New_Log(), Add_Line_Log() and Print_Log() will
        /// work only if writeLog is set to true.
        /// </summary>
        /// <param name="on"></param>
        static public void Set_Write_Log(bool on)
        {
            writeLog = on;
        }

        //set use log switch
        static public bool Get_Write_Log()
        {
            return writeLog;
        }

        //reset calculation log
        static public void New_Log()
        {
            if (writeLog)
                calcLog = "";
        }

        /// <summary>
        /// Add a line calculation log.
        /// You don't need to give a new line.
        /// Log must be initialized with New_Log()!
        /// </summary>
        /// <param name="line"></param>
        static public void Add_Line_Log(string line)
        {
            if (writeLog)
            {
                if (calcLog == "")
                    calcLog = line;
                else
                    calcLog = calcLog + System.Environment.NewLine + line;
            }
        }

        /// <summary>
        /// print calculation log to file
        /// </summary>
        /// <param name="path"></param>
        static public void Print_Log(string path)
        {
            if (writeLog)
                System.IO.File.WriteAllText(path, calcLog);
        }

        //print whole element list
        static public string Element_List_To_String(List<CalculationElement> elements)
        {
            string ret = "";

            //beginning brackets
            ret += Copy_N_Times("(", elements[0].NestedLevel);

            for (int n = 0; n < elements.Count; ++n)
            {
                ret = ret + elements[n].ToString();
                //if nested level changes print corresponding brackets
                if (n < elements.Count - 1)
                {
                    int nestedLevelChange = elements[n + 1].NestedLevel - elements[n].NestedLevel;
                    //nested level increases
                    if (nestedLevelChange > 0)
                        ret += " " + Copy_N_Times("(", nestedLevelChange);
                    //nested level decreases
                    else if (nestedLevelChange < 0)
                        ret += Copy_N_Times(")", Math.Abs(nestedLevelChange)) + " ";
                    else
                        ret += " ";
                }
            }

            //print missing brackets at the end
            ret += Copy_N_Times(")", elements[elements.Count - 1].NestedLevel);
            return ret;
        }

        //print whole element list
        static public string Element_List_Type_To_String(List<CalculationElement> elements)
        {
            string ret = "";

            //beginning brackets
            ret += Copy_N_Times("(", elements[0].NestedLevel);

            //elements and 
            for (int n = 0; n < elements.Count; ++n)
            {
                ret = ret + elements[n].TypeToString();
                //if nested level changes print corresponding brackets
                if (n < elements.Count - 1)
                {
                    int nestedLevelChange = elements[n + 1].NestedLevel - elements[n].NestedLevel;
                    //nested level increases
                    if (nestedLevelChange > 0)
                        ret += " " + Copy_N_Times("(", nestedLevelChange);
                    //nested level decreases
                    else if (nestedLevelChange < 0)
                        ret += Copy_N_Times(")", Math.Abs(nestedLevelChange)) + " ";
                    else
                        ret += " ";
                }
            }

            //print missing brackets at the end
            ret += Copy_N_Times(")", elements[elements.Count - 1].NestedLevel);
            return ret;
        }

        //Concatenates copyStr n times and returns the result.
        static public string Copy_N_Times(string copyStr, int times)
        {
            string ret = "";
            for (int n = 0; n < times; ++n)
            {
                ret = ret + copyStr;
            }
            return ret;
        }
        #endregion

        #region Result_Formatting

        /// <summary>
        /// Set number of decimals displayed by formatted calculation result.
        /// NOTE: Due to the limited accurracy of the double datatype the valid 
        /// area is 0 to 15. Any invalid input will rise an error.
        /// </summary>
        /// <param name="newFormatStr"></param>
        static public void Set_Num_Decimals(int newNumDec)
        { 
            if (newNumDec < 0 || newNumDec > 15)
                throw new Exception("Invalid number of decimals for number formatting.");
            numRoundDecimals = newNumDec;
        }

        /// <summary>
        /// Get number of decimals.
        /// </summary>
        /// <returns></returns>
        static public int Get_Num_Decimals()
        { return numRoundDecimals; }

        /// <summary>
        /// Set output format.
        /// </summary>
        /// <param name="newOF"></param>
        static public void Set_Output_Format(OutputFormat newOF)
        { outputFormat = newOF; }

        /// <summary>
        /// Get exponential view.
        /// </summary>
        /// <returns></returns>
        static public OutputFormat Get_OutputFormat()
        { return outputFormat; }
        #endregion

        /// <summary>
        /// Give line of script. Return type of control structure in this line.
        /// </summary>
        /// <param name="inStr"></param>
        /// <returns></returns>
        static private CtrlStructType CheckForCtrlStruct(string inStr)
        {
            inStr = inStr.Trim().ToLower();
            if (inStr.StartsWith(WHILE, StringComparison.CurrentCultureIgnoreCase))
                return CtrlStructType.While;
            else if (inStr.StartsWith(BREAK, StringComparison.CurrentCultureIgnoreCase))
                return CtrlStructType.Break;
            else if (inStr.StartsWith(CONTINUE, StringComparison.CurrentCultureIgnoreCase))
                return CtrlStructType.Continue;
            else if (inStr.StartsWith(LOOP, StringComparison.CurrentCultureIgnoreCase))
                return CtrlStructType.Loop;
            else if (inStr.StartsWith(IF, StringComparison.CurrentCultureIgnoreCase))
                return CtrlStructType.IF;
            else if (inStr.StartsWith(ELSEIF, StringComparison.CurrentCultureIgnoreCase))
                return CtrlStructType.ElseIf;
            else if (inStr.StartsWith(ELSE, StringComparison.CurrentCultureIgnoreCase))
                return CtrlStructType.Else;
            else if (inStr.StartsWith(ENDIF, StringComparison.CurrentCultureIgnoreCase))
                return CtrlStructType.EndIf;
            else
                return CtrlStructType.None;
        }

        /// <summary>
        /// Give script lines and current line. Receive line number of corresponding endif.
        /// </summary>
        /// <param name="inStr"></param>
        /// <param name="currLine"></param>
        /// <returns></returns>
        static private int FindEndOfIf(string[] inStr, int currLineNum)
        {
            int nestedIfs = 0;
            int endLine = currLineNum + 1;

            for (; endLine < inStr.Length; ++endLine)
            {
                string currLine = inStr[endLine].Trim().ToLower();
                if (currLine.StartsWith(IF, StringComparison.CurrentCultureIgnoreCase))
                    ++nestedIfs;
                else if (currLine.StartsWith(ENDIF, StringComparison.CurrentCultureIgnoreCase))
                {
                    if (nestedIfs == 0)
                        return endLine;
                    else
                        --nestedIfs;
                }
            }

            throw new Exception("EndIf missing");
        }

        /// <summary>
        /// Give script lines and current line. Receive line number of corresponding next if element.
        /// </summary>
        /// <param name="inStr"></param>
        /// <param name="currLine"></param>
        /// <returns></returns>
        static private int FindNextIfEl(string[] inStr, int currLineNum)
        {
            int nestedIfs = 0;
            int endLine = currLineNum + 1;

            for (; endLine < inStr.Length; ++endLine)
            {
                string currLine = inStr[endLine].Trim().ToLower();
                if (currLine.StartsWith(IF, StringComparison.CurrentCultureIgnoreCase))
                    ++nestedIfs;
                else if (currLine.StartsWith(ENDIF, StringComparison.CurrentCultureIgnoreCase))
                {
                    if (nestedIfs == 0)
                        return endLine;
                    else
                        --nestedIfs;
                }
                else if (currLine.StartsWith(ELSEIF, StringComparison.CurrentCultureIgnoreCase))
                {
                    if (nestedIfs == 0)
                        return endLine;
                }
                else if (currLine.StartsWith(ELSE, StringComparison.CurrentCultureIgnoreCase))
                {
                    if (nestedIfs == 0)
                        return endLine;
                }
            }

            throw new Exception("EndIf missing");
        }

        /// <summary>
        /// Give script. Receive script with added jumpmarks at if elements.
        /// </summary>
        /// <param name="inStr"></param>
        /// <returns></returns>
        static private string[] AddIfJumpMarks(string[] inStr)
        {
            int nestedIfs = 0;
            List<string> script = new List<string>(inStr);

            for (int endLine = 0; endLine < script.Count; ++endLine)
            {
                string currLine = script[endLine].Trim().ToLower();
                if (currLine.StartsWith(IF, StringComparison.CurrentCultureIgnoreCase))
                    ++nestedIfs;
                else if (currLine.StartsWith(ELSEIF, StringComparison.CurrentCultureIgnoreCase))
                {
                    script.Insert(endLine, "GOTO jmIF" + nestedIfs);
                    ++endLine;
                }
                else if (currLine.StartsWith(ELSE, StringComparison.CurrentCultureIgnoreCase))
                {
                    script.Insert(endLine, "GOTO jmIF" + nestedIfs);
                    ++endLine;
                }
                else if (currLine.StartsWith(ENDIF, StringComparison.CurrentCultureIgnoreCase))
                {
                    script.Insert(endLine, "JUMPMARK jmIF" + nestedIfs);
                    --nestedIfs;
                    ++endLine;
                }
            }

            return script.ToArray();
        }

        /// <summary>
        /// Give script lines and current line. Receive line number of corresponding while end.
        /// </summary>
        /// <param name="inStr"></param>
        /// <param name="currLine"></param>
        /// <returns></returns>
        static private int FindEndOfWhile(string[] inStr, int currLineNum)
        {
            int nestedWhiles = 0;
            int endLine = currLineNum + 1;

            for (; endLine < inStr.Length; ++endLine)
            {
                string currLine = inStr[endLine].Trim().ToLower();
                if (currLine.StartsWith(WHILE, StringComparison.CurrentCultureIgnoreCase))
                    ++nestedWhiles;
                else if (currLine.StartsWith(LOOP, StringComparison.CurrentCultureIgnoreCase))
                {
                    if (nestedWhiles == 0)
                        return endLine;
                    else
                        --nestedWhiles;
                }
            }

            throw new Exception("Loop missing");
        }

        /// <summary>
        /// Give script lines and current line. Receive line number of corresponding while beginning.
        /// </summary>
        /// <param name="inStr"></param>
        /// <param name="currLine"></param>
        /// <returns></returns>
        static private int FindBeginningOfWhile(string[] inStr, int currLineNum)
        {
            int nestedWhiles = 0;
            int endLine = currLineNum - 1;

            for (; endLine < inStr.Length; --endLine)
            {
                string currLine = inStr[endLine].Trim().ToLower();
                if (currLine.StartsWith(LOOP, StringComparison.CurrentCultureIgnoreCase))
                    ++nestedWhiles;
                else if (currLine.StartsWith(WHILE, StringComparison.CurrentCultureIgnoreCase))
                {
                    if (nestedWhiles == 0)
                        return endLine;
                    else
                        --nestedWhiles;
                }
            }

            throw new Exception("Loop missing");
        }

        /// <summary>
        /// Give script lines, current line and jump mark (without "jumpmark"-tag). Receive line number of jump mark.
        /// </summary>
        /// <param name="inStr"></param>
        /// <param name="currLineNum"></param>
        /// <param name="jumpMark"></param>
        /// <returns></returns>
        static private int FindJumpMark(string[] inStr, int currLineNum, string jumpMark)
        {
            int endLine = currLineNum - 1;

            for (; endLine < inStr.Length; ++endLine)
            {
                string currLine = inStr[endLine].Trim().ToLower();
                if (currLine.StartsWith("jumpMark", StringComparison.CurrentCultureIgnoreCase))
                {
                    currLine = currLine.Substring("jumpmark".Length).Trim();
                    if (currLine.StartsWith(jumpMark, StringComparison.CurrentCultureIgnoreCase))
                        return endLine;
                }
            }

            throw new Exception("Jump mark missing");
        }

        /// <summary>
        /// Give resultstring, receive interpretation as true or false.
        /// </summary>
        /// <param name="inStr"></param>
        /// <returns></returns>
        static private bool StrEqualsTrue(string inStr)
        {
            if (inStr == "0" || inStr == "0b00000000" || inStr == "0o0" || 
                inStr == "0x00" || inStr == "false")
                return false;
            else
                return true;
        }

        /// <summary>
        /// Execute several lines of a script.
        /// </summary>
        /// <param name="inStr"></param>
        /// <returns></returns>
        static public List<OutputElement> Exec_Script(string[] inStr)
        {
            // against endless loops
            DateTime startTime = DateTime.Now;
            DateTime endTime;
            double ENDLESS_LOOP_THRESHOLD = 5.0;    // 5 seconds

            List<OutputElement> results = new List<OutputElement>();

            inStr = AddIfJumpMarks(inStr);

            for (int n = 0; n <= inStr.GetUpperBound(0); ++n)
            {
                string line = inStr[n].Trim();
                OutputElement oEl;
                CtrlStructType ct;

                try
                {
                    line = Refine_Calc_String(line).Trim();
                }
                catch (Exception e)
                {
                    results.Add(new OutputElement(OutputElement.OutputType.Error, e.Message));
                    break;
                }

                //check for jumpmarks
                if (line.StartsWith("goto", StringComparison.CurrentCultureIgnoreCase))
                {
                    string jm = line.Substring("goto".Length).Trim();
                    try
                    {
                        n = FindJumpMark(inStr, n, jm);
                        continue;
                    }
                    catch (Exception)
                    {
                        results.Add(new OutputElement(OutputElement.OutputType.Error, "Jumpmark missing: " + jm));
                        break;
                    }
                }
                else if (line.StartsWith("jumpmark", StringComparison.CurrentCultureIgnoreCase))
                {
                    continue;
                }

                ct = CheckForCtrlStruct(inStr[n]);
                if (ct == CtrlStructType.IF)
                {
                    // calculate condition
                    oEl = CalcEngine.Calc_Main(line.Substring("if".Length));
                    if (oEl.GetOutputType() == OutputElement.OutputType.Error)
                    {
                        results.Add(oEl);
                    }
                    else if (oEl.GetOutputType() == OutputElement.OutputType.CalcOutput)
                    {
                        string res = oEl.GetMsg();
                        if (!StrEqualsTrue(res))
                        {
                            try
                            {
                                n = FindNextIfEl(inStr, n);
                            }
                            catch (Exception)
                            {
                                results.Add(new OutputElement(OutputElement.OutputType.Error, "Endif missing"));
                                break;
                            }
                        }
                    }
                    else
                    {
                        results.Add(new OutputElement(OutputElement.OutputType.Error, "Invalid condition for: " + line));
                        break;
                    }
                }
                else if (ct == CtrlStructType.ElseIf)
                {
                    oEl = CalcEngine.Calc_Main(line.Substring("elseif".Length));
                    if (oEl.GetOutputType() == OutputElement.OutputType.Error)
                    {
                        results.Add(oEl);
                    }
                    else if (oEl.GetOutputType() == OutputElement.OutputType.CalcOutput)
                    {
                        string res = oEl.GetMsg();
                        if (!StrEqualsTrue(res))
                        {
                            try
                            {
                                n = FindNextIfEl(inStr, n);
                            }
                            catch (Exception)
                            {
                                results.Add(new OutputElement(OutputElement.OutputType.Error, "Endif missing"));
                                break;
                            }
                        }
                    }
                }
                else if (ct == CtrlStructType.Else)
                {
                }
                else if (ct == CtrlStructType.EndIf)
                {
                }
                else if (ct == CtrlStructType.While)
                {
                    // calculate condition
                    oEl = CalcEngine.Calc_Main(line.Substring("while".Length));
                    if (oEl.GetOutputType() == OutputElement.OutputType.Error)
                    {
                        results.Add(oEl);
                    }
                    else if (oEl.GetOutputType() == OutputElement.OutputType.CalcOutput)
                    {
                        string res = oEl.GetMsg();
                        if (!StrEqualsTrue(res))
                        {
                            try
                            {
                                n = FindEndOfWhile(inStr, n);
                            }
                            catch (Exception)
                            {
                                results.Add(new OutputElement(OutputElement.OutputType.Error, "While without loop"));
                                continue;
                            }
                        }
                    }
                }
                else if (ct == CtrlStructType.Break)
                {
                    try
                    {
                        n = FindEndOfWhile(inStr, n);
                    }
                    catch (Exception)
                    {
                        results.Add(new OutputElement(OutputElement.OutputType.Error, "Break without loop"));
                        continue;
                    }
                }
                else if (ct == CtrlStructType.Continue)
                {
                    try
                    {
                        n = FindBeginningOfWhile(inStr, n) - 1;
                    }
                    catch (Exception)
                    {
                        results.Add(new OutputElement(OutputElement.OutputType.Error, "Continue without while"));
                        continue;
                    }
                }
                // increment cycle counter, execute script line after loop
                // and jump to while
                else if (ct == CtrlStructType.Loop)
                {
                    try
                    {
                        n = FindBeginningOfWhile(inStr, n) - 1;
                    }
                    catch (Exception)
                    {
                        results.Add(new OutputElement(OutputElement.OutputType.Error, "Loop without while"));
                        continue;
                    }
                }
                else if (ct == CtrlStructType.None)
                {

                    //perform calculation if no control structure
                    oEl = CalcEngine.Calc_Main(line);

                    // mark script input as script msg if successfully executed
                    if (oEl.GetOutputType() != OutputElement.OutputType.ScriptMSG)
                    {
                        results.Add(new OutputElement(OutputElement.OutputType.CalcInput, line));
                        results.Add(oEl);
                    }
                    else
                    {
                        results.Add(new OutputElement(OutputElement.OutputType.ScriptInput, line));
                        results.Add(new OutputElement(OutputElement.OutputType.ScriptMSG, oEl.GetMsg()));
                    }

                }


                // ask after endless loop threshold time,
                // whether the calculation should be stopped 
                endTime = DateTime.Now;
                if (endTime.Subtract(startTime) >= TimeSpan.FromSeconds(ENDLESS_LOOP_THRESHOLD))
                {
                    System.Windows.Forms.DialogResult res = System.Windows.Forms.MessageBox.Show("The calculation time exceeds the threshold for endless loops." +
                        System.Environment.NewLine + "Do you want to stop this calculation?", "Endless loop warning",
                        System.Windows.Forms.MessageBoxButtons.YesNo,
                        System.Windows.Forms.MessageBoxIcon.Information, System.Windows.Forms.MessageBoxDefaultButton.Button2);
                    if (res == System.Windows.Forms.DialogResult.Yes)
                    {
                        results.Add(new OutputElement(OutputElement.OutputType.Error, "Calculation stopped"));
                        break;
                    }
                    else
                        startTime = DateTime.Now;
                }
            }

            return results;
        }

        /// <summary>
        /// Give line of calculation as string return result.
        /// All possible error sources are catched.
        /// Executed script lines will return an empty string.
        /// The caller should ignore empty string returns.
        /// </summary>
        /// <param name="InputString"></param>
        /// <returns></returns>
        static public OutputElement Calc_Main(string InputString)
        {
            List<CalculationElement> list;
            string refinedInput;
            CalculationElement calcRes;

            double res;
            string resStr;

            New_Log();
            //write current state into log
            Add_Line_Log("Input: " + InputString);

            try
            {
                refinedInput = Refine_Calc_String(InputString).Trim();
            }
            catch (Exception e)
            {
                return new OutputElement(OutputElement.OutputType.Error, e.Message);
            }

            //write current state into log
            Add_Line_Log("Refined input: " + refinedInput);

            if (refinedInput == "")
                return new OutputElement(OutputElement.OutputType.Nothing, "");

            if (CalcScript.Is_Script_Fct(refinedInput))
            {
                try
                {
                    if (refinedInput.StartsWith("#echo", StringComparison.CurrentCultureIgnoreCase))
                        return new OutputElement(OutputElement.OutputType.ScriptMSG, refinedInput.Substring("#echo".Length).Trim());

                    bool scriptRes;
                    scriptRes = CalcScript.Execute_Script(refinedInput);
                    if (scriptRes == true)
                        return new OutputElement(OutputElement.OutputType.ScriptMSG, "");
                    else
                        return new OutputElement(OutputElement.OutputType.Error, "Script execution failed: " + refinedInput);
                }
                catch (Exception e)
                {
                    return new OutputElement(OutputElement.OutputType.Error, e.Message);
                }
            }


            //assignments must be catched here to get all elements and brackets
            string result;
            try
            {
                result = Do_Assignment(refinedInput);
                if (result != "")
                    return new OutputElement(OutputElement.OutputType.CalcOutput, result);

                list = Str_To_CalcElement_List(refinedInput);

                //write current state into log
                Add_Line_Log("Calculation Elements: " + Element_List_To_String(list));

                Assign_Unary_Sign(list);

                //write current state into log
                Add_Line_Log("Assigned unary minus: " + Element_List_To_String(list));

                //insert multiplication signs
                Add_Multiplication_Signs(list);

                //write current state into log
                Add_Line_Log("Added multiplication signs: " + Element_List_To_String(list));
                Add_Line_Log("Calc Types: " + Element_List_Type_To_String(list));

                //Calculate
                calcRes = Calc_Recursion(list);

                res = calcRes.GetValue();

                //set ans var
                CalcVariables.Set_Var("ans", res);

                resStr = formatOutput(res);
            }
            catch (Exception e)
            {
                return new OutputElement(OutputElement.OutputType.Error, e.Message);
            }

            return new OutputElement(OutputElement.OutputType.CalcOutput, resStr);
        }

        static private string formatOutput(double toFormat)
        {
            string formatStr;
            string resStr;

            formatStr = CalcEngine.Copy_N_Times("#", numRoundDecimals);
            // create output format string
            if (outputFormat == OutputFormat.Exponential)
            {
                formatStr = "0." + formatStr + "E+0";
                //convert result to string using format string and culture settings for decimal sign
                resStr = toFormat.ToString(formatStr, CalcEngine.Get_Decimal_Sign().NumberFormat);
            }
            else if (outputFormat == OutputFormat.Decimal)
            {
                formatStr = "#,##0." + formatStr;
                //convert result to string using format string and culture settings for decimal sign
                resStr = toFormat.ToString(formatStr, CalcEngine.Get_Decimal_Sign().NumberFormat);
            }
            else if (outputFormat == OutputFormat.Binary)
            {
                // convert double to integer for bin, oct and hex
                int ires = System.Convert.ToInt32(Math.Round(toFormat));
                resStr = Convert.ToString(ires, 2);

                // add preceding zeros until number of digits is multiple of 8
                while (resStr.Length % 8 != 0)
                    resStr = "0" + resStr;

                // add literal beginning
                resStr = "0b" + resStr;
            }
            else if (outputFormat == OutputFormat.Octal)
            {
                // convert double to integer for bin, oct and hex
                int ires = System.Convert.ToInt32(Math.Round(toFormat));
                resStr = Convert.ToString(ires, 8);

                // add literal beginning
                resStr = "0o" + resStr;
            }
            else if (outputFormat == OutputFormat.Hexadecimal)
            {
                // convert double to integer for bin, oct and hex
                int ires = System.Convert.ToInt32(Math.Round(toFormat));
                resStr = Convert.ToString(ires, 16).ToUpper();

                //add preceding zeros if number of digits is not straight
                if (resStr.Length % 2 != 0)
                    resStr = "0" + resStr;

                // add literal beginning
                resStr = "0x" + resStr;
            }
            else
                throw new Exception("Invalid output format");

            return resStr;
        }

        //give calculation string, return list of calcElements
        static public List<CalculationElement> Str_To_CalcElement_List(string InputStr)
        {
            string tempStr;
            List<CalculationElement> calcElements = new List<CalculationElement>();  //contains all elements
            int nestedLevel = 0;

            //copy inputstring into tempstring
            tempStr = string.Copy(InputStr);

            //split line into elements and save them in arrays
            while (tempStr != "")
            {
                CalcObjectInfo.objectInfo tempObject = CalcObjectInfo.Get_Object(tempStr, 0);
                if (tempObject.type == CalcObjectInfo.State.NotFound)
                    throw new Exception("Unknown element at: " + tempStr);
                else if (tempObject.type == CalcObjectInfo.State.OpenBracket)
                {
                    ++nestedLevel;
                    tempStr = tempStr.Substring(1).TrimStart();
                }
                else if (tempObject.type == CalcObjectInfo.State.CloseBracket)
                {
                    --nestedLevel;
                    tempStr = tempStr.Substring(1).TrimStart();
                }
                else
                {
                    CalculationElement tempEl;
                    //interprete current object
                    tempEl = CalcObjectInfo.ObjectInfo_To_CalcElement(tempObject);
                    tempEl.NestedLevel = nestedLevel;
                    calcElements.Add(tempEl);
                    tempStr = tempStr.Substring(tempObject.objectStr.Length).TrimStart();
                }
            }

            return calcElements;
        }

        //give function expression and parameter name, return list of of calcElements
        static public List<CalculationElement> Str_To_FctExpression_List(string InputStr, string param)
        {
            string tempStr;
            List<CalculationElement> calcElements = new List<CalculationElement>();  //contains all elements
            int nestedLevel = 0;

            //set parameter as variable that it can be found at string to calcElement conversion
            CalcVariables.Set_Var(param, 0.0);

            //copy inputstring into tempstring
            tempStr = string.Copy(InputStr);

            //split line into elements and save them in arrays
            while (tempStr != "")
            {
                CalcObjectInfo.objectInfo tempObject = CalcObjectInfo.Get_Object(tempStr, 0);
                //parameter might be unknown or known as variable
                if (tempObject.type == CalcObjectInfo.State.NotFound && tempStr.StartsWith(param) || 
                    tempObject.objectStr == param)
                {
                    calcElements.Add(new CalculationElement(CalculationElement.ElementType.Parameter, 0, nestedLevel));
                    tempStr = tempStr.Substring(param.Length).TrimStart();
                }
                //unknown object
                else if (tempObject.type == CalcObjectInfo.State.NotFound)
                    throw new Exception("Unknown element at: " + tempStr);
                //open bracket
                else if (tempObject.type == CalcObjectInfo.State.OpenBracket)
                {
                    ++nestedLevel;
                    tempStr = tempStr.Substring(1).TrimStart();
                }
                //close bracket
                else if (tempObject.type == CalcObjectInfo.State.CloseBracket)
                {
                    --nestedLevel;
                    tempStr = tempStr.Substring(1).TrimStart();
                }
                //usual calcElements
                else
                {
                    CalculationElement tempEl;
                    //interprete current object
                    tempEl = CalcObjectInfo.ObjectInfo_To_CalcElement(tempObject);
                    tempEl.NestedLevel = nestedLevel;
                    calcElements.Add(tempEl);
                    tempStr = tempStr.Substring(tempObject.objectStr.Length).TrimStart();
                }
            }

            return calcElements;
        }

        //add multiplication signs
        static public void Add_Multiplication_Signs(List<CalculationElement> objects)
        {
            //example: 3pi => 3 * pi
            for (int i = 0; i < objects.Count - 1; ++i)
            {
                //if number or var before number, var, fct
                //then connect via multiplication
                if ((objects[i].Type == CalculationElement.ElementType.Number || objects[i].Type == CalculationElement.ElementType.Variable) &&
                    (objects[i + 1].Type == CalculationElement.ElementType.Number || objects[i + 1].Type == CalculationElement.ElementType.Variable ||
                    objects[i + 1].Type == CalculationElement.ElementType.Function || objects[i + 1].Type == CalculationElement.ElementType.Parameter))
                {
                    int nestedLevel;
                    if (objects[i].NestedLevel <= objects[i + 1].NestedLevel)
                        nestedLevel = objects[i].NestedLevel;
                    else
                        nestedLevel = objects[i + 1].NestedLevel;
                    //new multiplication sign gets the same nested level as the foregoing element
                    CalculationElement tempEl = new CalculationElement(CalculationElement.ElementType.Operator, (int)CalcOperators.opIndex.mult, nestedLevel);
                    objects.Insert(i + 1, tempEl);
                }
                else if (i < objects.Count - 2)
                {
                    //treat 3 ^ -(4+2) like 3 ^ (-1 * (4+2))  but  keep 1-(20/16)
                    if (objects[i].Type == CalculationElement.ElementType.Operator && objects[i].Index == (int)CalcOperators.opIndex.pow &&                //operator
                        (objects[i + 1].Type == CalculationElement.ElementType.Operator && objects[i + 1].Index == (int)CalcOperators.opIndex.minus ||     //minus
                        objects[i + 1].Type == CalculationElement.ElementType.Operator && objects[i + 1].Index == (int)CalcOperators.opIndex.plus) &&      //or plus
                        (objects[i + 2].Type == CalculationElement.ElementType.Number || objects[i + 2].Type == CalculationElement.ElementType.Variable ||     //before
                        objects[i + 2].Type == CalculationElement.ElementType.Function || objects[i + 2].Type == CalculationElement.ElementType.Parameter) &&  //any kind of value
                        objects[i + 1].NestedLevel < objects[i + 2].NestedLevel)                                                        //with higher nested level
                    {
                        //Add brackets
                        int startNestedLevel = objects[i + 2].NestedLevel;
                        //get end of this nested level area
                        int endOfThisNestedLevel = i + 2;
                        while (objects[endOfThisNestedLevel].NestedLevel >= startNestedLevel)
                        {
                            ++endOfThisNestedLevel;
                            if (endOfThisNestedLevel == objects.Count)
                                break;
                        }
                        //increase nested level (add brackets around the whole stuff)
                        for (int n = i + 1; n < endOfThisNestedLevel; ++n)
                        {
                            objects[n] = new CalculationElement(objects[n], objects[n].NestedLevel + 1);
                        }

                        //Add new elements
                        //add 1
                        double sign = 1.0;
                        if (objects[i + 1].Index == (int)CalcOperators.opIndex.minus)
                            sign = -1.0;
                        CalculationElement tempEl = new CalculationElement(sign, objects[i + 1].NestedLevel);
                        objects.Insert(i + 2, tempEl);
                        //add multiplication sign
                        tempEl = new CalculationElement(CalculationElement.ElementType.Operator,
                            (int)CalcOperators.opIndex.mult, objects[i + 1].NestedLevel);
                        objects.Insert(i + 3, tempEl);
                        objects.RemoveAt(i + 1);
                    }
                }
            }
        }

        /// <summary>
        /// Assign unary sign to following number.
        /// Note: Does only work with numbers, not with variables or functions.
        /// </summary>
        /// <param name="list"></param>
        static public void Assign_Unary_Sign(List<CalculationElement> list)
        {
            //if operation begins with + or -
            if (list.Count > 1)
            {
                //unary sign must be within the same brackets
                if (list[0].NestedLevel == list[1].NestedLevel)
                {
                    if (list[0].Type == CalculationElement.ElementType.Operator)
                    {
                        if (list[0].Index == (int)CalcOperators.opIndex.minus)
                        {
                            list[1] = list[1].Negate();
                            list.RemoveAt(0);
                        }
                        else if (list[0].Index == (int)CalcOperators.opIndex.plus)
                            list.RemoveAt(0);
                        else
                            throw new Exception("2 operators in a row.");
                    }
                }
            }

            for (int n = 1; n < list.Count - 2; ++n)
            {
                //care for 3 ^ -(5 ^ 2)
                //unary sign must be within the same brackets
                if (list[n + 1].NestedLevel == list[n + 2].NestedLevel)
                {
                    //check for unary minus after operator
                    //treat 2*-3 like 2*(-3)  //;  f -3 like f (-3)
                    if (list[n].Type == CalculationElement.ElementType.Operator && list[n + 1].Type == CalculationElement.ElementType.Operator)
                    {
                        if (list[n + 1].Index == (int)CalcOperators.opIndex.minus)
                        {
                            if (list[n + 2].Type == CalculationElement.ElementType.Operator || list[n + 2].Type == CalculationElement.ElementType.Invalid)
                                throw new Exception("Unary sign before invalid operand");
                            else
                            {
                                list[n + 2] = list[n + 2].Negate();
                                list.RemoveAt(n + 1);
                            }
                        }
                        else if (list[n + 1].Index == (int)CalcOperators.opIndex.plus)
                            list.RemoveAt(n + 1);
                        else
                            throw new Exception("2 operators on after the other.");
                    }
                    if (list[n].Type == CalculationElement.ElementType.Function && list[n + 1].Type == CalculationElement.ElementType.Operator)
                    {
                        if (list[n + 1].Index == (int)CalcOperators.opIndex.minus)
                        {
                            if (list[n + 2].Type == CalculationElement.ElementType.Operator || list[n + 2].Type == CalculationElement.ElementType.Invalid)
                                throw new Exception("Unary sign before invalid operand");
                            else
                            {
                                list[n + 2] = list[n + 2].Negate();
                                list.RemoveAt(n + 1);
                            }
                        }
                        else if (list[n + 1].Index == (int)CalcOperators.opIndex.plus)
                            list.RemoveAt(n + 1);
                    }
                }
            }
        }

        //calc operators
        static public void Calc_Operations(List<CalculationElement> list)
        {
            //you need at least 3 elements for 1 operation or 2 elements for function and argument
            while (list.Count > 2)
            {
                int highestOrderVal = -1;                                   //highest operator order value
                int highOpIndex = -1;								        //index of highest order value

                //get operator with highest order value
                for (int i = 1; i < list.Count; ++i)
                {
                    if (list[i].Type == CalculationElement.ElementType.Operator)
                    {
                        int currentOpOrderVal = -1;
                        currentOpOrderVal = CalcOperators.Get_Operator_Order_Value(list[i]);
                        
                        System.Diagnostics.Debug.Assert(currentOpOrderVal > -1);
                        if (currentOpOrderVal > highestOrderVal)            //found new highest order value
                        {
                            highestOrderVal = currentOpOrderVal;
                            highOpIndex = i;
                        }
                    }
                }

                //if operator is 1st (0) or last element an error occurred
                if (highOpIndex < 1 || highOpIndex >= list.Count)
                    throw new Exception("Operand missing");

                try
                {
                    list[highOpIndex - 1] = CalcOperators.Simple_Calculation(list[highOpIndex - 1],
                        list[highOpIndex], list[highOpIndex + 1]);
                }
                catch (Exception e)
                {
                    throw new Exception(e.Message);
                }
                //delete other elements of this calculation
                list.RemoveAt(highOpIndex);
                list.RemoveAt(highOpIndex);

                //write current state into log
                Add_Line_Log("Calculated Operator: " + Element_List_To_String(list));
            }
        }

        //give whole list, return result
        static public CalculationElement Calc_Recursion(List<CalculationElement> list)
        {
            //Calc_Element_List
            List<CalculationElement> tempList = new List<CalculationElement>();
            CalculationElement tempEl;

            while (list.Count > 1)
            {
                int currentNestedLevel = -1;							    //nestedlevel (in how many brackets)
                int highestNestedLevel = -1;                                //highest nested level
                int highestNestedLevelStart = -1;                           //start of highest nested level area
                int highestNestedLevelLen = -1;                             //end of highest nested level area
                //still in the same area as the beginning of the highest nested level start
                bool atAStretch = false;
                tempList.Clear();
                //test: f(t)=5*delta(t)-1*(sigma(t)-sigma(t-5))
                //get area of highest nested level
                for (int i = 0; i < list.Count; ++i)
                {
                    //actualize current nested level on change of nestedlevel
                    if (list[i].NestedLevel > currentNestedLevel)
                    {
                        currentNestedLevel = list[i].NestedLevel;
                        //if found higher nested level
                        if (currentNestedLevel > highestNestedLevel)
                        {
                            highestNestedLevelStart = i;
                            highestNestedLevel = currentNestedLevel;
                            atAStretch = true;
                        }
                    }
                    else if (list[i].NestedLevel < currentNestedLevel)
                    {
                        currentNestedLevel = list[i].NestedLevel;
                        //end of highest nestedlevel
                        if (currentNestedLevel < highestNestedLevel && atAStretch == true)
                        {
                            highestNestedLevelLen = i - highestNestedLevelStart;
                            atAStretch = false;
                        }
                    }
                }

                //if no open brackets found calculate complete list
                if (highestNestedLevelStart == -1)
                {
                    highestNestedLevelStart = 0;
                    highestNestedLevelLen = list.Count;
                }
                
                //if no close bracket found (last element has nestedlevel > 0) 
                //or if no element after closing bracket of highest nested level area calculate rest of list
                if (highestNestedLevelLen == -1 || atAStretch)
                    highestNestedLevelLen = list.Count - highestNestedLevelStart;

                //close bracket found, open bracket not found is NOT catched here (bracket missing)

                //write current state into log
                Add_Line_Log(System.Environment.NewLine + "Solve bracket:");

                //copy to temporary list
                //Note: highestNestedLevelLen is 1 too high => Number of elements
                tempList.AddRange(list.GetRange(highestNestedLevelStart, highestNestedLevelLen));

                //calc area
                tempEl = Calc_Element_List(tempList);

                //replace calculated area by result
                list.RemoveRange(highestNestedLevelStart, highestNestedLevelLen);
                tempEl.NestedLevel = tempEl.NestedLevel - 1;
                list.Insert(highestNestedLevelStart, tempEl);

                //add multiplication signs around calculated bracket
                if (highestNestedLevelStart > 0)                    //at left side of bracket
                {
                    if (list[highestNestedLevelStart - 1].Type == CalculationElement.ElementType.Variable ||
                        list[highestNestedLevelStart - 1].Type == CalculationElement.ElementType.Parameter ||
                        list[highestNestedLevelStart - 1].Type == CalculationElement.ElementType.Number)
                        list.Insert(highestNestedLevelStart - 1, new CalculationElement(CalculationElement.ElementType.Operator,
                            (int)CalcOperators.opIndex.mult, list[highestNestedLevelStart - 1].NestedLevel));
                }
                if (list.Count - highestNestedLevelStart > 1)       //at right side of bracket
                {
                    if (list[highestNestedLevelStart + 1].Type == CalculationElement.ElementType.Variable ||
                        list[highestNestedLevelStart + 1].Type == CalculationElement.ElementType.Parameter ||
                        list[highestNestedLevelStart + 1].Type == CalculationElement.ElementType.Number)
                        list.Insert(highestNestedLevelStart - 1, new CalculationElement(CalculationElement.ElementType.Operator,
                            (int)CalcOperators.opIndex.mult, list[highestNestedLevelStart + 1].NestedLevel));
                }

                //write current state into log
                Add_Line_Log("Solve bracket finished" + System.Environment.NewLine);
                Add_Line_Log("Solved Bracket: " + Element_List_To_String(list));
            }

            list[0] = list[0].Interprete_Variable();

            return list[0];
        }

        /// <summary>
        /// Perform input checks: Trim, remove comments, and check if all brackets are closed.
        /// </summary>
        /// <param name="InputString"></param>
        /// <returns></returns>
        static public string Refine_Calc_String(string InputString) 
	    {
		    string tempStr;									        //contains changes of inputstring and used for return
            int pos = 0;

		    // copy inputstring into tempstring
            tempStr = string.Copy(InputString);

            // remove comments
            // "//" begins a comment if there is no escape character in front of it
            pos = tempStr.IndexOf(CalcGlobals.COMMENT);
            if (pos > 0)
            {
                if (tempStr[pos - 1] != CalcGlobals.ESCAPE_CHAR)
                    tempStr = tempStr.Remove(pos);
            }
            else if (pos == 0)
                tempStr = tempStr.Remove(pos);
            tempStr = tempStr.Trim(' ');

		    //skip if no input
            if (tempStr.Equals(""))
			    return "";

            //check if number of open and close brackets is equal
            if (!CalcObjectInfo.All_Brackets_Closed(InputString))
                throw new Exception("Bracket not closed");

            //allow ² and ³
            tempStr = tempStr.Replace("²", "^2");
            tempStr = tempStr.Replace("³", "^3");

            //allow ¼ ½ ¾...
            tempStr = tempStr.Replace("¼", "0" + CalcEngine.Get_Decimal_Sign().NumberFormat.NumberDecimalSeparator + "25");
            tempStr = tempStr.Replace("½", "0" + CalcEngine.Get_Decimal_Sign().NumberFormat.NumberDecimalSeparator + "5");
            tempStr = tempStr.Replace("¾", "0" + CalcEngine.Get_Decimal_Sign().NumberFormat.NumberDecimalSeparator + "75");

            //replace % by /100
            tempStr = tempStr.Replace("%", "/100");
            tempStr = tempStr.Replace("‰", "/1000");

            //allow {[]}
            tempStr = tempStr.Replace("{", "(");
            tempStr = tempStr.Replace("[", "(");
            tempStr = tempStr.Replace("}", ")");
            tempStr = tempStr.Replace("]", ")");

            // allow true and false
            while(true)
            {
                // replace true
                int boolPos = tempStr.IndexOf("true", StringComparison.CurrentCultureIgnoreCase);
                if (boolPos > -1)
                {
                    tempStr = tempStr.Remove(boolPos, "true".Length);
                    tempStr = tempStr.Insert(boolPos, "1");
                    continue;
                }

                // replace false
                boolPos = tempStr.IndexOf("false", StringComparison.CurrentCultureIgnoreCase);
                if (boolPos > -1)
                {
                    tempStr = tempStr.Remove(boolPos, "false".Length);
                    tempStr = tempStr.Insert(boolPos, "0");
                    continue;
                }

                if (boolPos == -1)
                    break;
            }

            tempStr = Convert_BinOctHexToDec(tempStr);

            return tempStr;
	    }

        /// <summary>
        /// Searches a string for Binary, Octal or Hexadecimal numbers
        /// and convert them to decimal.
        /// </summary>
        /// <param name="inStr"></param>
        /// <returns></returns>
        static public string Convert_BinOctHexToDec(string inStr)
        {
            const string HexStarter = "0x";
            const string OctStarter = "0o";
            const string BinStarter = "0b";
            
            int startHex = 0;
            int startOct = 0;
            int startBin = 0;

            while (true)
            {
                int startPos;
                if (startHex < inStr.Length)
                    startPos = inStr.IndexOf(HexStarter, startHex, StringComparison.CurrentCultureIgnoreCase);
                else
                    startPos = -1;
                // if Hex found
                if (startPos > -1)
                {
                    int lastPos = startPos + 2;
                    while (lastPos < inStr.Length)
                    {
                        // if between '0' and '9' or 'A' and 'F'
                        if ((inStr[lastPos] >= '0' && inStr[lastPos] <= '9') || (inStr[lastPos] >= 'a' && inStr[lastPos] <= 'f')
                             || (inStr[lastPos] >= 'A' && inStr[lastPos] <= 'F'))
                            ++lastPos;
                        // literal is finished
                        else
                            break;
                    }

                    // convert only if more than beginning
                    if (lastPos > startPos + 2)
                    {
                        // convert hex literal to dec integer
                        string hexLiteral = inStr.Substring(startPos + 2, lastPos - startPos - 2);
                        int decNum = System.Convert.ToInt32(hexLiteral, 16);
                        //replace hex with dec
                        inStr = inStr.Substring(0, startPos) + decNum.ToString() + inStr.Substring(lastPos);
                        startHex = 0;
                        continue;
                    }
                    else
                        startHex = startPos + 1;
                }

                if (startOct < inStr.Length)
                    startPos = inStr.IndexOf(OctStarter, startOct, StringComparison.CurrentCultureIgnoreCase);
                else
                    startPos = -1;
                // if oct found
                if (startPos > -1)
                {
                    int lastPos = startPos + 2;
                    while (lastPos < inStr.Length)
                    {
                        // if between '0' and '7'
                        if (inStr[lastPos] >= '0' && inStr[lastPos] <= '7')
                            ++lastPos;
                        // literal is finished
                        else
                            break;
                    }

                    // convert only if more than beginning
                    if (lastPos > startPos + 2)
                    {
                        // convert oct literal to dec integer
                        string octLiteral = inStr.Substring(startPos + 2, lastPos - startPos - 2);
                        int decNum = System.Convert.ToInt32(octLiteral, 8);
                        //replace oct with dec
                        inStr = inStr.Substring(0, startPos) + decNum.ToString() + inStr.Substring(lastPos);
                        startOct = 0;
                        continue;
                    }
                    else
                        startOct = startPos + 1;
                }

                if (startBin < inStr.Length)
                    startPos = inStr.IndexOf(BinStarter, startBin, StringComparison.CurrentCultureIgnoreCase);
                else
                    startPos = -1;
                // if bin found
                if (startPos > -1)
                {
                    int lastPos = startPos + 2;
                    while (lastPos < inStr.Length)
                    {
                        // if between '0' or '1'
                        if (inStr[lastPos] == '0' || inStr[lastPos] == '1')
                            ++lastPos;
                        // literal is finished
                        else
                            break;
                    }

                    // convert only if more than beginning
                    if (lastPos > startPos + 2)
                    {
                        // convert bin literal to dec integer
                        string binLiteral = inStr.Substring(startPos + 2, lastPos - startPos - 2);
                        int decNum = System.Convert.ToInt32(binLiteral, 2);
                        //replace hex with dec
                        inStr = inStr.Substring(0, startPos) + decNum.ToString() + inStr.Substring(lastPos);
                        startBin = 0;
                        continue;
                    }
                    else
                        startBin = startPos + 1;
                }

                // no more literals found
                if (startPos == -1)
                    break;
            }

            return inStr;
        }

        static public CalculationElement Calc_Element_List(List<CalculationElement> list)
        {
            try
            {
                //calc functions from right to left
                CalcFunctions.Calc_Functions(list);

                //write current state into log
                Add_Line_Log("Calculated functions: " + Element_List_To_String(list));

                //calculate operations
                Calc_Operations(list);

                //write current state into log
                Add_Line_Log("Calculated operations: " + Element_List_To_String(list));
            }
            catch (Exception e)
            {
                throw new Exception(e.Message);
            } 

            if (list.Count > 1)
                throw new Exception("Watch the syntax");

            return list[0];
        }

        #region Angle_Mode
        //give angle mode; save in anglemode
        static public void Set_Angle_Mode(angleModes IAngleMode)
	    {
		    angleMode = IAngleMode;
	    }

	    //get anglemode
        static public angleModes Get_Angle_Mode()
	    {
		    return angleMode;
	    }
        #endregion

        #region DecimalSign
        static void Default_Culture()
        {
            if (CultureInfo.CurrentCulture.NumberFormat.NumberDecimalSeparator == ",")
                currCult = DecimalSign.comma;
            else
                currCult = DecimalSign.point;
        }

        static public void Set_Decimal_Sign(DecimalSign dSign)
        {
            currCult = dSign;
        }

        static public CultureInfo Get_Decimal_Sign()
        {
            if (currCult == DecimalSign.comma)
                return cultures[1];
            else
                return cultures[0];
        }
        #endregion

        #region Assignment

        /// <summary>
        /// Give assignment string, returns variable or function prints.
        /// Returns "" if there is no assignment.
        /// Throw exceptions if anything wrong happens.
        /// </summary>
        /// <param name="inputStr"></param>
        /// <returns></returns>
        static private string Do_Assignment(string inputStr)
        {
            int equalPos = inputStr.IndexOf('=');

            // skip if there is no equal sign
            if (equalPos < 0)
                return "";

            // skip if it is part of another operator
            for (int i = 0; i <= CalcOperators.opCompareStr.GetUpperBound(0); ++i)
            {
                int thisOpPos = -1;
                thisOpPos = inputStr.IndexOf(CalcOperators.opCompareStr[i]);
                if (thisOpPos > -1)
                {
                    if (equalPos >= thisOpPos && equalPos <= thisOpPos + CalcOperators.opCompareStr[i].Length)
                        return "";
                }
            }

            //get position of brackets in function definition
            int openBracketPos = inputStr.Remove(equalPos).IndexOf("(");
            int closeBracketPos = inputStr.Remove(equalPos).IndexOf(")");

            //function definition
            if (openBracketPos > -1 && closeBracketPos > -1)
            {
                if (openBracketPos > closeBracketPos)
                    throw new Exception("Wrong bracket order");

                string fctDef = inputStr.Substring(0, equalPos);
                string fctExpr = inputStr.Substring(equalPos + 1);

                //exception if variable with same name exists
                string name = fctDef.Substring(0, fctDef.IndexOf("("));
                if (CalcVariables.Get_Var_Index(name) > -1)
                    throw new Exception("\"" + name + "\" is already used as a variable");

                try
                {
                    return CalcFunctions.Set_Function(fctDef, fctExpr);
                }
                catch (Exception e)
                {
                    throw new Exception(e.Message);
                }
            }
            else if (openBracketPos == -1 ^ closeBracketPos == -1)
            {
                throw new Exception("Bracket missing");
            }
            else        //var definition
            {
                string varName = inputStr.Substring(0, equalPos).Trim();

                //exception if variable with same name exists
                if (CalcFunctions.Get_Fct_Index(varName) > -1)
                    throw new Exception("\"" + varName + "\" is already used as a function");

                bool convertRes = false;
                string valString;
                double varExpr;
                try
                {
                    valString = Refine_Calc_String(inputStr.Substring(equalPos + 1).Trim());
                    convertRes = Double.TryParse(Calc_Main(valString).GetMsg(), NumberStyles.AllowExponent |
                                                 NumberStyles.AllowThousands | NumberStyles.Number, 
                                                 Get_Decimal_Sign().NumberFormat, out varExpr);
                    CalcVariables.Set_Var(varName, varExpr);

                    return varName + " = " + CalcVariables.Get_Var(varName);
                }
                catch (Exception e)
                {
                    if (convertRes)
                        throw new Exception(e.Message);
                    else
                        throw new Exception("Invalid assignment: " + inputStr);
                }
            }
        }

        #endregion
    }
}
