﻿/*
 * ##############################################
 * CalcFunctions
 * Main part of the QuickCalculator
 * ##############################################
 * 
 * You may copy and distribute this code and the 
 * whole program as you want.
 * This code is not meant to be sold.
 * 
 * Written by Philipp Erler
 * Contact me for feedback, new version or other causes:
 * ph.erler@gmx.net
 * 
 * Version info:
 * Last change: 2012 | 04 | 29
 * .net-Framework: V. 4.0
 * 
 */

using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace CalculationE
{
    static class CalcFunctions
    {
        private enum StdFcts                        {  sin,   cos,   tan,   asin,   acos,   atan,   sinh,   cosh,   tanh,   abs, 
                                                       sigma,   delta,   sinc,   fact,   ln,   lg,   ld,  sqrt, not, neg};

        static public string[] stdFunctions =       { "sin", "cos", "tan", "asin", "acos", "atan", "sinh", "cosh", "tanh", "abs", 
                                                      "sigma", "delta", "sinc", "fact", "ln", "lg", "ld", "sqrt", "!", "~" };

        private const int MAX_Functs = 50;

        public struct function                                         //input:   f(x) = 3 sin x
        {
            public string name;                                         //example: f
            public CalculationElement parameter;             //example: x
            public List<CalculationElement> expression;      //example: 3 sin x

            public function(string name, CalculationElement parameter, List<CalculationElement> expression)
            {
                this.name = name;
                this.parameter = parameter;
                this.expression = expression;
            }

            public bool Is_Std_Fct()
            {
                for (int i = 0; i < stdFunctions.Count(); ++i)
                {
                    if (this.name == stdFunctions[i])
                        return true;
                }
                return false;
            }
        }

        static private List<function> calcFcts = new List<function>();

        //constructor to set standard functions
        static CalcFunctions()
        {
            //set standard functions
            for (int n = 0; n < stdFunctions.Count(); ++n)
            {
                Set_Std_Function((StdFcts)n);
            }
        }

        static bool Is_Std_Fct(function fct)
        {
            for (int i = 0; i < stdFunctions.Count(); ++i)
            {
                if (fct.name == stdFunctions[i])
                    return true;
            }
            return false;
        }

        static public bool Is_Std_Fct(int fctIndex)
        {
            for (int i = 0; i < stdFunctions.Count(); ++i)
            {
                if (calcFcts[fctIndex].name == stdFunctions[i])
                    return true;
            }
            return false;
        }

        static bool Is_Std_Fct(string fctName)
        {
            for (int i = 0; i < stdFunctions.Count(); ++i)
            {
                if (fctName == stdFunctions[i])
                    return true;
            }
            return false;
        }

        //give functiondef string (example: f(x)) and function expression string (example: 2x + 5)
        //return success message
        //Note: Overwrites existing function with the same name. Do NOT use while calculation is running!
        static public string Set_Function(string functionDef, string functionExpr)
        {
            int posOpenBracket = -1, posCloseBracket = -1;
            int functionIndex = -1;

            //refine definition and expression
            functionDef = functionDef.Trim();
            functionExpr = functionExpr.Trim();

            //throw error if something is missing
            if (functionDef == "")
                throw new Exception("Missing function definition");
            else if (functionExpr == "")
                throw new Exception("Missing function expression");

            //get bracket positions
            posOpenBracket = functionDef.IndexOf('(');
            posCloseBracket = functionDef.IndexOf(')');

            //get parts of new function
            string name = functionDef.Remove(posOpenBracket).ToLower().Trim();              //TODO: without lowercase
            string parameterName = functionDef.Remove(0, posOpenBracket + 1).Remove(posCloseBracket - posOpenBracket - 1).ToLower();
            string expression = functionExpr.ToLower();

            //skip if parameter is missing
            if (posOpenBracket == -1 || posCloseBracket == -1 || Math.Abs(posCloseBracket - posOpenBracket) == 1 ||
                posOpenBracket > posCloseBracket || posCloseBracket == 0)
                throw new Exception("Missing function parameter");

            //already used?
            for (int i = 0; i <= calcFcts.Count - 1; ++i)
            {
                if (calcFcts[i].name == name)
                {
                    //prevent overwriting of std_fct
                    if (Is_Std_Fct(name))
                        return Print_Function(i);
                    functionIndex = i;
                    break;
                }
            }

            //standard parameter
            CalculationElement parameter = new CalculationElement(CalculationElement.ElementType.Parameter, 0, 0);

            //contains something like sin x
            List<CalculationElement> expList = CalcEngine.Str_To_FctExpression_List(expression, parameterName);

            //CalcEngine.Assign_Unary_Sign(expList);
            //CalcEngine.Add_Multiplication_Signs(expList);
            //CalcEngine.Assign_Unary_Sign(expList);

            //add new function
            if (functionIndex == -1)
            {
                calcFcts.Add(new function(name, parameter, expList));
                functionIndex = calcFcts.Count - 1;
            }
            else
                calcFcts[functionIndex] = new function(name, parameter, expList);

            return Print_Function(functionIndex);
        }

        //give functiondef string (example: f(x)) and stc function type
        //Note: Overwrites existing function with the same name. Do NOT use while calculation is running!
        static private void Set_Std_Function(StdFcts fct)
        {
            //set function properties
            string name = stdFunctions[(int)fct];

            //skip if already set
            for (int i = 0; i <= calcFcts.Count - 1; ++i)
            {
                if (calcFcts[i].name == name)
                    return;
            }

            //TODO: 1 standard parameter for alle 1-parametric functions
            //standard parameter
            CalculationElement parameter = new CalculationElement(CalculationElement.ElementType.Parameter, 0, 0);

            //contains something like sin x
            List<CalculationElement> expList = new List<CalculationElement>();

            //add reference to this future function entry
            expList.Add(new CalculationElement(CalculationElement.ElementType.Function, calcFcts.Count, 0));
            //add parameter
            expList.Add(parameter);

            calcFcts.Add(new function(name, parameter, expList));
        }

        /// <summary>
        /// Returns number of assigned functions
        /// </summary>
        /// <returns></returns>
        static public int Count()
        { 
            return calcFcts.Count(); 
        }

        static public void Del_Functions()
        {
            for (int i = 0; i < calcFcts.Count(); ++i)
            {
                if (!calcFcts[i].Is_Std_Fct())                                 //do not delete std fct
                    calcFcts.RemoveAt(i);
            }
        }

        //remove calc function
        static public void Del_Function(int index)
        {
            if (index < 0 || index > calcFcts.Count - 1)
                return;                                                             //skip if invalid index
            else
            {
                if (!calcFcts[index].Is_Std_Fct())                             //do not delete std fct
                    calcFcts.RemoveAt(index);
            }

        }

        //return list of all functions
        static public string Print_Functions()
        {
            string fctList = "Function list:";
            for (int i = 0; i <= calcFcts.Count - 1; ++i)
            {
                fctList = fctList + System.Environment.NewLine + Print_Function(i);
            }

            return fctList;
        }

        //return list of all user functions
        static public string Print_User_Functions()
        {
            List<string> userFcts = Get_User_Functions();

            return "Function list:" + System.Environment.NewLine + 
                CalcScript.Concat_String_Array(userFcts.ToArray(), System.Environment.NewLine);
        }

        /// <summary>
        /// Return list of all user functions.
        /// </summary>
        /// <returns></returns>
        static public List<string> Get_User_Functions()
        {
            List<string> fctList = new List<string>();
            string fctPrint;

            for (int i = 0; i <= calcFcts.Count - 1; ++i)
            {
                if (!Is_Std_Fct(calcFcts[i]))
                {
                    fctPrint = calcFcts[i].name + "(" + calcFcts[i].parameter.ToString() + ") = " + 
                        CalcEngine.Element_List_To_String(calcFcts[i].expression);
                    fctList.Add(fctPrint);
                }
            }

            return fctList;
        }

        /// <summary>
        /// Return list of all user functions' names.
        /// </summary>
        /// <returns></returns>
        static public List<string> Get_User_Function_Names()
        {
            List<string> fctList = new List<string>();
            string fctPrint;

            for (int i = 0; i <= calcFcts.Count - 1; ++i)
            {
                if (!Is_Std_Fct(calcFcts[i]))
                {
                    fctPrint = calcFcts[i].name;
                    fctList.Add(fctPrint);
                }
            }

            return fctList;
        }

        /// <summary>
        /// Return list of all user functions' indiezes.
        /// </summary>
        /// <returns></returns>
        static public List<int> Get_User_Function_Indiezes()
        {
            List<int> fctList = new List<int>();

            for (int i = 0; i <= calcFcts.Count - 1; ++i)
            {
                if (!Is_Std_Fct(calcFcts[i]))
                {
                    fctList.Add(i);
                }
            }

            return fctList;
        }

        //return function name, brackets, parameter and expression
        static public string Print_Function(int FunctionIndex)
        {
            if (FunctionIndex < 0 || FunctionIndex > calcFcts.Count - 1)
                return "";

            string fctNameValue = calcFcts[FunctionIndex].name + "(" + calcFcts[FunctionIndex].parameter.ToString() +
                                    ") = " + CalcEngine.Element_List_To_String(calcFcts[FunctionIndex].expression);

            return fctNameValue;
        }

        /* outdated! Use calcengine.element_list_to_string() instead
        //return string of an expression calcElement list
        static private string Print_Fct_Expression(List<CalcEngine.CalculationElement> list)
        {
            string res = "";
            for (int n = 0; n < list.Count; ++n)
            {
                res += list[n].ToString();
                if (n < list.Count - 1)
                    res += " ";
            }
            return res;
        }
        */

        static public string Get_Function_Name(int Index)
        { return calcFcts[Index].name; }

        static public string Get_Function_Parameter(int Index)
        { return CalcVariables.Get_Var_Name(calcFcts[Index].parameter.Index); }

        /// <summary>
        /// Get calc fct index. Return -1 if not found.
        /// </summary>
        /// <param name="name"></param>
        /// <returns></returns>
        static public int Get_Fct_Index(string name)
        {
            int fctIndex = -1;

            name = name.Trim().ToLower();

            for (int i = 0; i <= calcFcts.Count - 1; ++i)
            {
                if (calcFcts[i].name == name)
                {
                    fctIndex = i;
                    break;
                }
            }
                
            return fctIndex;
        }
            
        /// <summary>
        /// Get calc fct index of parameter. Return -1 if not found.
        /// </summary>
        /// <param name="name"></param>
        /// <returns></returns>
        static public int Get_Fct_Param_Index(string name)
        {
            int fctIndex = -1;

            name = name.Trim().ToLower();

            if (name.StartsWith("x"))
            {
                name = name.Substring(1);
                bool res = int.TryParse(name, out fctIndex);
                if (res)
                    return fctIndex;
                else
                    return -1;
            }
                
            return fctIndex;
        }

        /// <summary>
        /// Calculate Function with given parameter
        /// </summary>
        /// <param name="functionIndex"></param>
        /// <param name="setInValue"></param>
        /// <returns></returns>
        static public CalculationElement Calc_Function(CalculationElement fct, double argVal)
        {
            //try standard functions
            CalculationElement calcEl = Try_Calc_Standard_Function(fct, argVal);
            if (calcEl.Type != CalculationElement.ElementType.Invalid)
                return calcEl;

            //treat as seperate calculation
            List<CalculationElement> tempList = new List<CalculationElement>();
            Copy_To_Temp_List(calcFcts[fct.Index].expression, tempList);
            Replace_Arg_By_Value(tempList, calcFcts[fct.Index].parameter, argVal);
            CalcEngine.Assign_Unary_Sign(tempList);
            CalcEngine.Add_Multiplication_Signs(tempList);
            CalculationElement tempEl = CalcEngine.Calc_Recursion(tempList); //recursion instead of calc_element_list
            if (fct.Sign() == -1)
                tempEl = tempEl.Negate();
            return tempEl;
        }

        /// <summary>
        /// Replace function argument in calcElement list by given value
        /// </summary>
        /// <param name="list"></param>
        /// <param name="arg"></param>
        /// <param name="val"></param>
        static private void Replace_Arg_By_Value(List<CalculationElement> list,
            CalculationElement arg, double val)
        {
            for (int n = 0; n < list.Count; ++n)
            {
                if (list[n].Type == arg.Type && list[n].Index == arg.Index)
                {
                    //same nestedlevel as replaced element
                    CalculationElement calcE = new CalculationElement(val, list[n].NestedLevel);
                    //negate if replaced parameter was negated
                    if (list[n].Sign() == -1)
                        calcE = calcE.Negate();
                    list.Insert(n, calcE);
                    list.RemoveAt(n + 1);
                }
            }
        }

        static private void Copy_To_Temp_List(List<CalculationElement> oldList,
            List<CalculationElement> destList)
        {
            for (int n = 0; n < oldList.Count; ++n)
            {
                destList.Add(oldList[n]);
            }
        }

        static public bool Element_Is_Standard_Function(CalculationElement el)
        {
            if (el.Type != CalculationElement.ElementType.Function)
                throw new Exception("No Function given");

            for (int n = 0; n < stdFunctions.Count(); ++n)
            {
                if (Get_Fct_Index(stdFunctions[n]) == el.Index)
                    return true;
            }

            return false;
        }

        /// <summary>
        /// Try to calculate standard functions. If an error occurs return an invalid CalcElement.
        /// </summary>
        /// <param name="fct"></param>
        /// <param name="arg"></param>
        /// <returns></returns>
        static CalculationElement Try_Calc_Standard_Function(CalculationElement fct, double arg)
        {
            double res;

            if (fct.Index == Get_Fct_Index(stdFunctions[(int)StdFcts.sin]))             // sin
            {
                if (CalcEngine.Get_Angle_Mode() == CalculationE.CalcEngine.angleModes.Degree)
                    arg = arg * Math.PI / 180.0;
                res = Math.Sin(arg);
            }
            else if (fct.Index == Get_Fct_Index(stdFunctions[(int)StdFcts.cos]))        // cos
            {
                if (CalcEngine.Get_Angle_Mode() == CalculationE.CalcEngine.angleModes.Degree)
                    arg = arg * Math.PI / 180.0;
                res = Math.Cos(arg);
            }
            else if (fct.Index == Get_Fct_Index(stdFunctions[(int)StdFcts.tan]))        // tan
            {
                if (CalcEngine.Get_Angle_Mode() == CalculationE.CalcEngine.angleModes.Degree)
                    arg = arg * Math.PI / 180.0;
                res = Math.Tan(arg);
            }
            else if (fct.Index == Get_Fct_Index(stdFunctions[(int)StdFcts.asin]))       // asin
            {
                if (CalcEngine.Get_Angle_Mode() == CalculationE.CalcEngine.angleModes.Degree)
                    arg = arg * Math.PI / 180.0;
                res = Math.Asin(arg);
            }
            else if (fct.Index == Get_Fct_Index(stdFunctions[(int)StdFcts.acos]))       // acos
            {
                if (CalcEngine.Get_Angle_Mode() == CalculationE.CalcEngine.angleModes.Degree)
                    arg = arg * Math.PI / 180.0;
                res = Math.Acos(arg);
            }
            else if (fct.Index == Get_Fct_Index(stdFunctions[(int)StdFcts.atan]))       // atan
            {
                if (CalcEngine.Get_Angle_Mode() == CalculationE.CalcEngine.angleModes.Degree)
                    arg = arg * Math.PI / 180.0;
                res = Math.Atan(arg);
            }
            else if (fct.Index == Get_Fct_Index(stdFunctions[(int)StdFcts.sinh]))       // sinh
            {
                if (CalcEngine.Get_Angle_Mode() == CalculationE.CalcEngine.angleModes.Degree)
                    arg = arg * Math.PI / 180.0;
                res = Math.Sinh(arg);
            }
            else if (fct.Index == Get_Fct_Index(stdFunctions[(int)StdFcts.cosh]))       // cosh
            {
                if (CalcEngine.Get_Angle_Mode() == CalculationE.CalcEngine.angleModes.Degree)
                    arg = arg * Math.PI / 180.0;
                res = Math.Cosh(arg);
            }
            else if (fct.Index == Get_Fct_Index(stdFunctions[(int)StdFcts.tanh]))       // tanh
            {
                if (CalcEngine.Get_Angle_Mode() == CalculationE.CalcEngine.angleModes.Degree)
                    arg = arg * Math.PI / 180.0;
                res = Math.Tanh(arg);
            }
            else if (fct.Index == Get_Fct_Index(stdFunctions[(int)StdFcts.abs]))        // abs
                res = Math.Abs(arg);
            else if (fct.Index == Get_Fct_Index(stdFunctions[(int)StdFcts.sinc]))       // sinc
            {
                if (arg == 0)
                    throw new Exception("Division by zero");
                res = Math.Sin(arg) / arg;
            }
            else if (fct.Index == Get_Fct_Index(stdFunctions[(int)StdFcts.delta]))      // delta
            {
                if (arg == 0)
                    res = 1;
                else
                    res = 0;
            }
            else if (fct.Index == Get_Fct_Index(stdFunctions[(int)StdFcts.sigma]))      // sigma
            {
                if (arg < 0)
                    res = 0;
                else
                    res = 1;
            }
            else if (fct.Index == Get_Fct_Index(stdFunctions[(int)StdFcts.fact]))       // factorial
            {
                Int64 intRes;
                Int64 intArg = Convert.ToInt32(Math.Round(arg, 0));
                if (intArg < 0)
                    throw new Exception("No factorial for negative numbers");
                intRes = Factorial(intArg);
                res = Convert.ToDouble(intRes);
            }
            else if (fct.Index == Get_Fct_Index(stdFunctions[(int)StdFcts.ln]))         // ln
            {
                res = Math.Log(arg, Math.E);
            }
            else if (fct.Index == Get_Fct_Index(stdFunctions[(int)StdFcts.lg]))         // lg
            {
                res = Math.Log10(arg);
            }
            else if (fct.Index == Get_Fct_Index(stdFunctions[(int)StdFcts.ld]))         // ld
            {
                res = Math.Log(arg, 2.0);
            }
            else if (fct.Index == Get_Fct_Index(stdFunctions[(int)StdFcts.sqrt]))       // sqrt
            {
                res = Math.Sqrt(arg);
            }
            else if (fct.Index == Get_Fct_Index(stdFunctions[(int)StdFcts.not]))        // not
            {
                if (arg == 0.0)
                    res = 1.0;
                else
                    res = 0.0;
            }
            else if (fct.Index == Get_Fct_Index(stdFunctions[(int)StdFcts.neg]))        // neg
            {
                Int64 iarg = System.Convert.ToInt64(Math.Round(arg));
                iarg = ~iarg;
                res = System.Convert.ToDouble(iarg);
            }
            else
                return new CalculationElement(CalculationElement.ElementType.Invalid, -1, -1);

            return new CalculationElement(res, fct.NestedLevel);
        }

        /// <summary>
        /// Calculate the factorial of an integer.
        /// Thanks to Rosetta-Code: http://rosettacode.org/wiki/Factorial
        /// </summary>
        /// <param name="number"></param>
        /// <returns></returns>
        static Int64 Factorial(Int64 number)
        {
            Int64 result = 1;
            for (Int64 factor = 2; factor <= number; factor++)
            {
                result *= factor;
            }
            return result;
        }

        /// <summary>
        /// Calculate all functions in list.
        /// Begin with function that has highest nested level. Calculate right to left.
        /// </summary>
        /// <param name="list"></param>
        static public void Calc_Functions(List<CalculationElement> list)
        {
            //exception if last element is a function
            if (list[list.Count - 1].Type == CalculationElement.ElementType.Function)
                throw new Exception("Argument missing");

            while (true)
            {
                //get highest nested level fct
                int highestIndex = -1;
                int highestNestedLevel = -1;

                for (int n = 0; n < list.Count - 1; ++n)
                {
                    //if found last highest safe index
                    if (list[n].NestedLevel >= highestNestedLevel && list[n].Type == CalculationElement.ElementType.Function) 
                    {
                        highestIndex = n;
                        highestNestedLevel = list[n].NestedLevel;
                    }
                }

                //if nothing found finish
                if (highestIndex == -1)
                    break;

                //assign unary plus and minus to argument
                if (list[highestIndex + 1].Type == CalculationElement.ElementType.Operator)
                {
                    //remove plus
                    if (list[highestIndex + 1].Index == (int)CalcOperators.opIndex.plus)
                        list.RemoveAt(highestIndex + 1);
                    //assign minus
                    else if (list[highestIndex + 1].Index == (int)CalcOperators.opIndex.minus)
                    {
                        //throw if nothing behind the unary sign
                        if (highestIndex + 2 >= list.Count)
                            throw new Exception("Argument missing");
                        list.RemoveAt(highestIndex + 1);
                        list[highestIndex + 1] = list[highestIndex + 1].Negate();
                    }
                }

                /*
                //if argument is no number
                if (list[highestIndex + 1].Type != CalcEngine.ElementType.Number)
                    throw new Exception("Invalid argument");
                */
                CalculationElement tempE;

                //Calc Function
                if (list[highestIndex + 1].Type == CalculationElement.ElementType.Invalid ||
                    list[highestIndex + 1].Type == CalculationElement.ElementType.Function ||
                    list[highestIndex + 1].Type == CalculationElement.ElementType.Operator)         //functions cannot be an argument
                    throw new Exception("Invalid argument at: " + list[highestIndex].ToString() + list[highestIndex + 1].ToString());
                else
                    tempE = Calc_Function(list[highestIndex], list[highestIndex + 1].Interprete_Variable().GetValue());

                //replace function and argument by result
                list.Insert(highestIndex, tempE);
                list.RemoveAt(highestIndex + 1);
                list.RemoveAt(highestIndex + 1);
            }
        }
    }
}
