﻿/*
 * ##############################################
 * Calculation Objects
 * This is the part that gives info about Object
 * (Length, Stringrepresentation, etc.) and 
 * turns the Calculation-string into CalcElements
 * In other words its the CalcString abstraction layer.
 * Main part of the QuickCalculator
 * ##############################################
 * 
 * You may copy and distribute this code and the 
 * whole program as you want.
 * This code is not meant to be sold.
 * 
 * Written by Philipp Erler
 * Contact me for feedback, new version or other causes:
 * ph.erler@gmx.net
 * 
 * Version info:
 * Last change: 2012 | 04 | 29
 * .net-Framework: V. 4.0
 * 
 */

using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Globalization;

namespace CalculationE
{
    static class CalcObjectInfo
    {

        /// <summary>
        /// state of entry enum 
        /// </summary>
        public enum State
        {
            Beginning, Operator, Dezimalsign, Thousandsign, Assignment, Function, Variable, Number,
            OpenBracket, CloseBracket, Exponent, Hex, Oct, Bin, NotFound
        };

        /// <summary>
        /// for search for operator, variable, function etc functions
        /// </summary>
        public struct objectInfo
        {
            public bool found;
            public int pos;
            public int len;
            public string objectStr;
            public State type;

            public objectInfo(bool found, int pos, int len, string objectStr, State type)
            {
                this.found = found;
                this.pos = pos;
                this.len = len;
                this.objectStr = objectStr;
                this.type = type;
            }

            /// <summary>
            /// use only for found = false
            /// </summary>
            /// <param name="found"></param>
            public objectInfo(bool found)
            {
                this.found = found;
                this.len = 0;
                this.pos = 0;
                this.objectStr = "";
                this.type = State.NotFound;
            }

            /// <summary>
            /// pos and type will not fit anymore
            /// </summary>
            /// <param name="objectStr"></param>
            public objectInfo(string objectStr)
            {
                this.objectStr = objectStr;
                this.found = true;
                this.pos = -1;
                this.len = objectStr.Length;
                this.type = State.NotFound;
            }
        }

        /// <summary>
        /// struct for all state infos
        /// </summary>
        public struct StateInfo
        {
            public bool bracketMissing;                                 //true if more open than close brackets
            public bool dezimalSign;                                    //true if dez sign is already set
            public bool withinObject;                                   //true if checked char is within an operator, value etc.
            public objectInfo thisObject;                               //info of current object
            public bool assignment;                                     //true if '=' occurs within the string
            public bool argument;                                       //true if it is an argument
        }

        //Parts of a double-containing string
        private enum DoubleParseState
        { start, sign, beforeDecimalSign, decimalSign, afterDecimalSign, exponentStart, exponentSign, exponentDigits };

        /// <summary>
        /// Convert objectInfo to CalcElement. Nestedlevel is set to -1.
        /// </summary>
        /// <param name="operand"></param>
        /// <returns></returns>
        static public CalculationElement ObjectInfo_To_CalcElement(objectInfo oInfo)
        {
            CalculationElement calcE = new CalculationElement(CalculationElement.ElementType.Invalid, -1, -1);

            oInfo.objectStr = oInfo.objectStr.Trim();

            if (oInfo.objectStr == "")
                return calcE;

            oInfo.objectStr = oInfo.objectStr.ToLower();

            //check for valid double
            {
                double d = 0.0;
                bool success;
                success = Double.TryParse(oInfo.objectStr, NumberStyles.AllowExponent | 
                    NumberStyles.AllowThousands | NumberStyles.Number, CalcEngine.Get_Decimal_Sign().NumberFormat, out d);
                if (success)
                {
                    calcE.SetValue(d);
                    calcE.Type = CalculationElement.ElementType.Number;
                    return calcE;
                }
            }

            //check for operator
            {
                int tempIndex;
                tempIndex = CalcOperators.Get_Operator_Index(oInfo.objectStr);
                if (tempIndex != -1)
                {
                    calcE.Index = tempIndex;
                    calcE.Type = CalculationElement.ElementType.Operator;
                    return calcE;
                }
            }

            //check for calc function
            {
                int tempIndex;
                tempIndex = CalcFunctions.Get_Fct_Index(oInfo.objectStr);
                if (tempIndex != -1)
                {
                    calcE.Index = tempIndex;
                    calcE.Type = CalculationElement.ElementType.Function;
                    return calcE;
                }
            }

            //check for variables
            {
                int tempIndex;
                tempIndex = CalcVariables.Get_Var_Index(oInfo.objectStr);
                if (tempIndex != -1)
                {
                    calcE.Index = tempIndex;
                    calcE.Type = CalculationElement.ElementType.Variable;
                    return calcE;
                }
            }

            throw new Exception("Unknown Object " + oInfo.objectStr);
        }

        /// <summary>
        /// return the nestedlevel of a specified char within a calculation expression
        /// pos is the position of the char; valid area is 0 smaller pos smaller inputStr.Length
        /// if char at this pos is "(" nestedLevel does not consider this sign
        /// </summary>
        /// <param name="inputStr"></param>
        /// <param name="pos"></param>
        /// <returns></returns>
        static public int Get_Nested_Level(string inputStr, int pos)
        {
            int nestedLevel = 0;

            if (pos < 0)
                throw new Exception(string.Format("Pos = {0}; Pos must be >= 0", pos));
            if (pos > inputStr.Length)
                throw new Exception(string.Format("Pos = {0}; Pos must be < inputStr.Length [{1}]", pos, inputStr.Length));

            for (int i = 0; i < pos; ++i)
            {
                if (inputStr.Substring(i, 1) == "(")
                    ++nestedLevel;
                else if (inputStr.Substring(i, 1) == ")")
                    --nestedLevel;
            }

            return nestedLevel;
        }

        /// <summary>
        /// get number of occurrences of 1 char within a string
        /// </summary>
        /// <param name="searchStr"></param>
        /// <param name="lookForThis"></param>
        /// <returns></returns>
        static public int Get_Num_Char_In_Str(string searchStr, char lookForThis)
        {
            int numberOfOccurrence = 0;

            for (int i = 0; i < searchStr.Length; ++i)
            {
                if (searchStr.Substring(i, 1) == lookForThis.ToString())
                    ++numberOfOccurrence;
            }

            return numberOfOccurrence;
        }

        //are all brackets in a string closed
        static public bool All_Brackets_Closed(string inputStr)
        {
            return (CalcObjectInfo.Get_Num_Char_In_Str(inputStr, '(') == CalcObjectInfo.Get_Num_Char_In_Str(inputStr, ')'));
        }

        /// <summary>
        /// Give string and pos, return if a bracket is missing at the position
        /// </summary>
        /// <param name="inputStr"></param>
        /// <param name="pos"></param>
        /// <returns></returns>
        static private bool Get_Bracket_State(string inputStr, int pos)
        {
            string leftPart = inputStr;
            if (pos + 1 < inputStr.Length)
                leftPart = inputStr.Remove(pos + 1);

            int numOpenBracket = Get_Num_Char_In_Str(leftPart, '(');
            int numCloseBracket = Get_Num_Char_In_Str(leftPart, ')');
            if (numOpenBracket > numCloseBracket)
                return true;
            else
                return false;

        }

        /// <summary>
        /// what meaning has the char after a specific position in a string
        /// </summary>
        /// <param name="inputStr"></param>
        /// <param name="pos"></param>
        /// <returns></returns>
        static public StateInfo Get_State_At_Pos(string inputStr, int pos)
        {
            StateInfo Info = new StateInfo();

            //if empty string return initialized info
            if (inputStr.Trim() == "")
            {
                Info.thisObject = new objectInfo(true, 0, 0, "", State.Beginning);
                Info.bracketMissing = false;
                Info.argument = false;
                Info.assignment = false;
                Info.dezimalSign = false;
                Info.withinObject = false;
                return Info;
            }

            //check for invalid pos
            if (pos < 0 || pos > inputStr.Length)
                throw new Exception("Position invalid");
            if (pos == inputStr.Length && pos > 0)
                pos--;

            //go back if target is space
            while (inputStr[pos] == ' ')
            {
                pos--;
                if (pos < 0)
                    throw new Exception("Position invalid");
            }

            //init info
            Info.argument = false;
            if (inputStr.IndexOf('=') > -1)
                Info.assignment = true;
            else
                Info.assignment = false;
            Info.bracketMissing = Get_Bracket_State(inputStr, pos);
            Info.dezimalSign = false;
            Info.withinObject = false;
            Info.thisObject = new objectInfo(false);

            //separate target object
            Info.thisObject = Separate_Object(inputStr, pos);

            //argument of function
            int assumedFunction = Info.thisObject.pos - 1;          //if after function and after spaces
            if (assumedFunction > -1)
            {
                for (; Is_Space_At_Pos(inputStr, assumedFunction).found; assumedFunction--)
                {
                    if (assumedFunction < 0)
                        break;
                }
                if (assumedFunction > -1)
                {
                    if (Is_User_Function_At_Pos(inputStr, assumedFunction).found)
                        Info.argument = true;
                    else                                            //if after function and between brackets
                    {
                        int functionPos = -1;
                        int functionLen = -1;
                        string beforePos = inputStr.Remove(pos).ToLower();
                        if (beforePos != "")
                        {
                            //get last function before pos
                            //user functions
                            for (int n = 0; n < CalcFunctions.Count(); ++n)
                            {
                                int thisFunctionPos = beforePos.LastIndexOf(CalcFunctions.Get_Function_Name(n));
                                if (thisFunctionPos > functionPos)
                                {
                                    functionLen = CalcFunctions.Get_Function_Name(n).Length;
                                    functionPos = thisFunctionPos;
                                }
                            }
                            //open bracket must be between function and pos
                            if (functionPos > -1)
                            {
                                int openBracketAfterFct = functionPos + functionLen + inputStr.Substring(functionPos + functionLen).IndexOf('(');
                                if (openBracketAfterFct > -1 && openBracketAfterFct < pos && openBracketAfterFct >= functionPos + functionLen)
                                {
                                    //between function and open bracket must only be white space
                                    if (inputStr.Substring(functionPos + functionLen, openBracketAfterFct - (functionPos + functionLen)) == "")
                                    {
                                        int nestedLevelOpenBracket = Get_Nested_Level(inputStr, openBracketAfterFct);
                                        int nextCloseBracket = openBracketAfterFct + 1;
                                        while (nextCloseBracket > -1)
                                        {
                                            if (nextCloseBracket < inputStr.Length - 1)
                                                nextCloseBracket = inputStr.IndexOf(')', nextCloseBracket + 1);
                                            else
                                            {
                                                nextCloseBracket = -1;
                                                break;
                                            }

                                            //if pos between open and close bracket
                                            if (Get_Nested_Level(inputStr, nextCloseBracket + 1) == nestedLevelOpenBracket && nextCloseBracket > pos)
                                                Info.argument = true;
                                            else
                                                break;
                                        }
                                        if (nextCloseBracket < 0)
                                            Info.argument = true;
                                    }
                                }
                            }
                        }
                    }
                }
            }

            //check for numbers / dezimal sign
            if (Info.thisObject.type == State.Number)
            {
                //dezimal
                Info.thisObject = Get_Number_At_Pos(inputStr, pos);
                if (Info.thisObject.objectStr.Contains(System.Globalization.CultureInfo.CurrentCulture.NumberFormat.NumberDecimalSeparator))
                    Info.dezimalSign = true;
            }

            if (pos != Info.thisObject.pos + Info.thisObject.len)
                Info.withinObject = true;

            return Info;
        }

        #region Object_Info_Etc

        /// <summary>
        /// give string and position, return object at position
        /// Note: Only for numbers, variables, functions
        /// separators are: blanks, brackets, operators, assignment
        /// Important: This function returns numbers and dezimal signs separated
        /// </summary>
        /// <param name="inputStr"></param>
        /// <param name="pos"></param>
        /// <returns></returns>
        static public objectInfo Separate_Object(string inputStr, int pos)
        {
            objectInfo thisObject = new objectInfo();

            //space
            thisObject = Is_Space_At_Pos(inputStr, pos);
            if (thisObject.found)
                return thisObject;

            //number
            thisObject = Is_Number_At_Pos(inputStr, pos);
            if (thisObject.found)
                return thisObject;

            //dezimal sign
            thisObject = Is_Dezimal_Sign_At_Pos(inputStr, pos);
            if (thisObject.found)
                return thisObject;

            //thousand sign
            thisObject = Is_Thousand_Sign_At_Pos(inputStr, pos);
            if (thisObject.found)
                return thisObject;

            //assignment
            thisObject = Is_Assignment_At_Pos(inputStr, pos);
            if (thisObject.found)
                return thisObject;

            //close bracket
            thisObject = Is_Close_Bracket_At_Pos(inputStr, pos);
            if (thisObject.found)
                return thisObject;

            //open bracket
            thisObject = Is_Open_Bracket_At_Pos(inputStr, pos);
            if (thisObject.found)
                return thisObject;

            //operator
            thisObject = Is_Operator_At_Pos(inputStr, pos);
            if (thisObject.found)
                return thisObject;

            //user function
            thisObject = Is_User_Function_At_Pos(inputStr, pos);
            if (thisObject.found)
                return thisObject;

            //std function
            thisObject = Is_Standard_Function_At_Pos(inputStr, pos);
            if (thisObject.found)
                return thisObject;

            //variable
            thisObject = Is_Variable_At_Pos(inputStr, pos);
            if (thisObject.found)
                return thisObject;

            return new objectInfo(false);
        }

        /// <summary>
        /// give string and position, return object at position
        /// Note: Only for numbers, variables, functions
        /// separators are: blanks, brackets, operators, assignment
        /// Important: This function returns complete numbers
        /// </summary>
        /// <param name="inputStr"></param>
        /// <param name="pos"></param>
        /// <returns></returns>
        static public objectInfo Get_Object(string inputStr, int pos)
        {
            objectInfo thisObject;

            //space
            thisObject = Is_Space_At_Pos(inputStr, pos);
            if (thisObject.found)
                return thisObject;

            //dezimal sign
            thisObject = Is_Dezimal_Sign_At_Pos(inputStr, pos);
            if (thisObject.found)
                return Get_Number_At_Pos(inputStr, pos);

            //thousand sign
            thisObject = Is_Thousand_Sign_At_Pos(inputStr, pos);
            if (thisObject.found)
                return Get_Number_At_Pos(inputStr, pos);

            /* not neccessary anymore
            //assignment
            thisObject = Is_Assignment_At_Pos(inputStr, pos);
            if (thisObject.found)
                return thisObject;
             */

            //close bracket
            thisObject = Is_Close_Bracket_At_Pos(inputStr, pos);
            if (thisObject.found)
                return thisObject;

            //open bracket
            thisObject = Is_Open_Bracket_At_Pos(inputStr, pos);
            if (thisObject.found)
                return thisObject;

            //operator
            thisObject = Is_Operator_At_Pos(inputStr, pos);
            if (thisObject.found)
                return thisObject;

            //user function
            thisObject = Is_User_Function_At_Pos(inputStr, pos);
            if (thisObject.found)
                return thisObject;

            //std function
            thisObject = Is_Standard_Function_At_Pos(inputStr, pos);
            if (thisObject.found)
                return thisObject;

            //user function parameter
            thisObject = Is_Fct_Parameter_At_Pos(inputStr, pos);
            if (thisObject.found)
                return thisObject;

            //variable
            thisObject = Is_Variable_At_Pos(inputStr, pos);
            if (thisObject.found)
                return thisObject;

            //number
            thisObject = Is_Number_At_Pos(inputStr, pos);
            if (thisObject.found)
                return Get_Number_At_Pos(inputStr, pos);

            return new objectInfo(false);
        }

        /// <summary>
        /// check if string can be converted into a double
        /// </summary>
        /// <param name="numberStr"></param>
        /// <returns></returns>
        static public bool Is_Numeric(string numberStr)
        {
            try
            {
                System.Convert.ToDouble(numberStr);
                return true;
            }
            catch (Exception)
            {
                return false;
            }
        }

        static public objectInfo Is_String_At_Pos(string inputStr, string compStr, int pos)
        {

            int foundPos = inputStr.IndexOf(compStr, StringComparison.CurrentCultureIgnoreCase);

            while (foundPos > -1)
            {
                if (pos >= foundPos && pos < foundPos + compStr.Length)
                    return new objectInfo(true, foundPos, compStr.Length, inputStr.Substring(foundPos, compStr.Length), State.NotFound);

                foundPos = inputStr.IndexOf(compStr, foundPos + 1);
            }
            return new objectInfo(false);
        }

        static public objectInfo Is_Space_At_Pos(string inputStr, int pos)
        {
            if (inputStr[pos] == ' ')
                return new objectInfo(true, pos, 1, inputStr.Substring(pos, 1), State.NotFound);
            else
                return new objectInfo(false);
        }

        static public objectInfo Is_Open_Bracket_At_Pos(string inputStr, int pos)
        {
            if (inputStr[pos] == '(')
                return new objectInfo(true, pos, 1, inputStr.Substring(pos, 1), State.OpenBracket);
            else
                return new objectInfo(false);
        }

        static public objectInfo Is_Close_Bracket_At_Pos(string inputStr, int pos)
        {
            if (inputStr[pos] == ')')
                return new objectInfo(true, pos, 1, inputStr.Substring(pos, 1), State.CloseBracket);
            else
                return new objectInfo(false);
        }

        static public objectInfo Is_Assignment_At_Pos(string inputStr, int pos)
        {
            if (inputStr[pos] == '=')
                return new objectInfo(true, pos, 1, inputStr.Substring(pos, 1), State.Assignment);
            else
                return new objectInfo(false);
        }

        static public objectInfo Is_Number_At_Pos(string inputStr, int pos)
        {
            if (inputStr[pos] >= '0' && inputStr[pos] <= '9')
                return new objectInfo(true, pos, 1, inputStr.Substring(pos, 1), State.Number);
            else
                return new objectInfo(false);
        }

        static public objectInfo Get_Double_At_Pos(string inputStr, int pos)
        {
            DoubleParseState state = DoubleParseState.start;
            int n;
            int doubleEnd = -1;
            bool valid = false;
            bool breakFor = false;
            string tempStr = inputStr.Substring(pos);
            int sinceLastGroupSep = -1;

            //check validity of input
            if (inputStr == null)
                return new objectInfo(false);
            if (inputStr.Trim() == "" || pos < 0 || pos > inputStr.Length)
                return new objectInfo(false);

            //use a state model for different steps of double validity check
            for (n = 0; n <= tempStr.Length; ++n)
            {
                switch (state)
                {
                    case DoubleParseState.start:
                        if (tempStr[n] == '+' || tempStr[n] == '-')             //allow sign
                            state = DoubleParseState.sign;
                        else if (Is_Number_At_Pos(tempStr, n).found)            //allow numbers
                        {
                            state = DoubleParseState.beforeDecimalSign;
                            valid = true;
                            doubleEnd = n;
                        }
                        else if (tempStr[n] == ' ')                             //ignore white space
                        { }                                                     //allow decimal sign
                        else if ((sinceLastGroupSep == 3 || sinceLastGroupSep == -1) && Is_Dezimal_Sign_At_Pos(tempStr, n).found)
                            state = DoubleParseState.decimalSign;               //allow thousands separator
                        else if ((sinceLastGroupSep == 3 || sinceLastGroupSep == -1) && Is_Thousand_Sign_At_Pos(tempStr, n).found)
                            sinceLastGroupSep = 0;
                        else
                            breakFor = true;
                        break;
                    case DoubleParseState.sign:
                        if (Is_Number_At_Pos(tempStr, n).found)                 //allow numbers
                        {
                            state = DoubleParseState.beforeDecimalSign;
                            valid = true;
                            doubleEnd = n;
                        }                                                       //allow decimal sign
                        else if ((sinceLastGroupSep == 3 || sinceLastGroupSep == -1) && Is_Dezimal_Sign_At_Pos(tempStr, n).found)
                            state = DoubleParseState.decimalSign;               //allow thousands separator
                        else if ((sinceLastGroupSep == 3 || sinceLastGroupSep == -1) && Is_Thousand_Sign_At_Pos(tempStr, n).found)
                            sinceLastGroupSep = 0;
                        else
                            breakFor = true;
                        break;
                    case DoubleParseState.beforeDecimalSign:
                        if (Is_Number_At_Pos(tempStr, n).found)                 //allow numbers
                        {
                            if (sinceLastGroupSep > -1)
                                ++sinceLastGroupSep;
                            valid = true;
                            doubleEnd = n;
                        }                                                       //allow number group separator
                        else if ((sinceLastGroupSep == 3 || sinceLastGroupSep == -1) && Is_Thousand_Sign_At_Pos(tempStr, n).found)
                        {
                            sinceLastGroupSep = 0;
                        }                                                       //allow decimal sign
                        else if ((sinceLastGroupSep == 3 || sinceLastGroupSep == -1) && Is_Dezimal_Sign_At_Pos(tempStr, n).found)
                        {
                            state = DoubleParseState.decimalSign;
                            valid = true;
                            doubleEnd = n;
                        }
                        else
                            breakFor = true;
                        break;
                    case DoubleParseState.decimalSign:
                        if (Is_Number_At_Pos(tempStr, n).found)                 //allow numbers
                        {
                            state = DoubleParseState.afterDecimalSign;
                            valid = true;
                            doubleEnd = n - 1;
                        }
                        else if (tempStr[n] == 'e' || tempStr[n] == 'E')        //allow exponent
                            state = DoubleParseState.exponentStart;
                        else
                            breakFor = true;
                        break;
                    case DoubleParseState.afterDecimalSign:
                        if (Is_Number_At_Pos(tempStr, n).found)                 //allow numbers
                        {
                            state = DoubleParseState.afterDecimalSign;
                            valid = true;
                            doubleEnd = n - 1;
                        }
                        else if (tempStr[n] == 'e' || tempStr[n] == 'E')        //allow exponent
                            state = DoubleParseState.exponentStart;
                        else
                            breakFor = true;
                        break;
                    case DoubleParseState.exponentStart:
                        if (tempStr[n] == '+' || tempStr[n] == '-')             //allow sign
                            state = DoubleParseState.exponentSign;
                        else if (Is_Number_At_Pos(tempStr, n).found)            //allow numbers
                        {
                            state = DoubleParseState.afterDecimalSign;
                            valid = true;
                            doubleEnd = n - 1;
                        }
                        else
                            breakFor = true;
                        break;
                    case DoubleParseState.exponentSign:
                        if (Is_Number_At_Pos(tempStr, n).found)                 //allow numbers
                        {
                            state = DoubleParseState.afterDecimalSign;
                            valid = true;
                            doubleEnd = n - 1;
                        }
                        else
                            breakFor = true;
                        break;
                    case DoubleParseState.exponentDigits:
                        if (Is_Number_At_Pos(tempStr, n).found)                 //allow numbers
                        {
                            state = DoubleParseState.afterDecimalSign;
                            valid = true;
                            doubleEnd = n - 1;
                        }
                        else
                            breakFor = true;
                        break;
                }

                if (breakFor)
                    break;

            }

            if (valid)
            {
                if (doubleEnd - pos < tempStr.Length)
                    return new objectInfo(true, pos, doubleEnd - pos, tempStr.Remove(doubleEnd - pos + 1), State.Number);
                else
                    return new objectInfo(true, pos, doubleEnd - pos, tempStr, State.Number);
            }
            else
                return new objectInfo(false);
        }

        static public objectInfo Get_Number_At_Pos(string inputStr, int pos)
        {
            if (Is_Number_At_Pos(inputStr, pos).found == false)
                return new objectInfo(false);

            int leftEnd = pos;
            int rightEnd;

            while (leftEnd > 0)
            {
                if (Is_Number_At_Pos(inputStr, leftEnd - 1).found || Is_Dezimal_Sign_At_Pos(inputStr, leftEnd - 1).found ||
                    Is_Thousand_Sign_At_Pos(inputStr, leftEnd - 1).found || Is_Exponent_Sign_At_Pos(inputStr, leftEnd - 1).found)
                    leftEnd--;
                else
                    break;
            }

            for (rightEnd = pos; rightEnd < inputStr.Length && (Is_Number_At_Pos(inputStr, rightEnd).found ||
                Is_Dezimal_Sign_At_Pos(inputStr, rightEnd).found || Is_Thousand_Sign_At_Pos(inputStr, rightEnd).found || 
                Is_Exponent_Sign_At_Pos(inputStr, rightEnd).found); ++rightEnd) ;

            return new objectInfo(true, leftEnd, rightEnd - leftEnd, inputStr.Substring(leftEnd, rightEnd - leftEnd), State.Number);
        }

        static public objectInfo Is_Dezimal_Sign_At_Pos(string inputStr, int pos)
        {
            string dec = CalcEngine.Get_Decimal_Sign().NumberFormat.NumberDecimalSeparator;
            if (inputStr[pos].ToString() == dec)
                return new objectInfo(true, pos, dec.Length, inputStr.Substring(pos, dec.Length), State.Dezimalsign);
            else
                return new objectInfo(false);
        }

        static public objectInfo Is_Thousand_Sign_At_Pos(string inputStr, int pos)
        {
            string grp = CalcEngine.Get_Decimal_Sign().NumberFormat.NumberGroupSeparator;
            if (inputStr[pos].ToString() == grp)
                return new objectInfo(true, pos, grp.Length, inputStr.Substring(pos, grp.Length), State.Thousandsign);
            else
                return new objectInfo(false);
        }

        static public objectInfo Is_Exponent_Sign_At_Pos(string inputStr, int pos)
        {
            //e or E must be followed by an integer (begins with +,-, or digit
            if (pos >= inputStr.Length - 1)             //no exponent if it is the last thing in line
                return new objectInfo(false);
            //e at pos
            if ((inputStr[pos] == 'e' || inputStr[pos] == 'E') && (Is_Number_At_Pos(inputStr, pos + 1).found || 
                inputStr[pos + 1] == '+' || inputStr[pos + 1] == '-'))
                return new objectInfo(true, pos, 1, inputStr.Substring(pos, 1), State.Exponent);
            //exp number at pos
            else if (pos > 0)
            {
                if ((inputStr[pos - 1] == 'e' || inputStr[pos - 1] == 'E') && (Is_Number_At_Pos(inputStr, pos).found ||
                inputStr[pos] == '+' || inputStr[pos] == '-'))
                    return new objectInfo(true, pos, 1, inputStr.Substring(pos, 1), State.Exponent);
                else
                    return new objectInfo(false);
            }
            else
                return new objectInfo(false);
        }

        static public objectInfo Is_Operator_At_Pos(string inputStr, int pos)
        {
            objectInfo temp = new objectInfo(false);
            for (int i = 0; i <= CalcOperators.opOrderVal.GetUpperBound(0); ++i)
            {
                objectInfo tempInfo = Is_String_At_Pos(inputStr, CalcOperators.opCompareStr[i], pos);
                if (tempInfo.found && tempInfo.len > temp.len)
                    temp = new objectInfo(true, tempInfo.pos, tempInfo.len, inputStr.Substring(tempInfo.pos, tempInfo.len), State.Operator);
            }
            return temp;
        }

        static public objectInfo Is_Standard_Function_At_Pos(string inputStr, int pos)
        {
            objectInfo temp = new objectInfo(false);
            //NOTE: If cos is checked before cosh, the result might be wrong
            for (int i = CalcFunctions.stdFunctions.GetUpperBound(0); i >= 0; i--)
            {
                objectInfo tempInfo = Is_String_At_Pos(inputStr, CalcFunctions.stdFunctions[i], pos);
                if (tempInfo.found && tempInfo.len > temp.len)
                    temp = new objectInfo(true, tempInfo.pos, tempInfo.len, inputStr.Substring(tempInfo.pos, tempInfo.len), State.Function);
            }
            return temp;
        }

        static public objectInfo Is_User_Function_At_Pos(string inputStr, int pos)
        {
            objectInfo temp = new objectInfo(false);
            for (int i = 0; i <= CalcFunctions.Count() - 1; ++i)
            {
                objectInfo tempInfo = Is_String_At_Pos(inputStr, CalcFunctions.Get_Function_Name(i), pos);
                if (tempInfo.found && tempInfo.len > temp.len)
                    temp = new objectInfo(true, tempInfo.pos, tempInfo.len, inputStr.Substring(tempInfo.pos, tempInfo.len), State.Function);
            }
            return temp;
        }

        static public objectInfo Is_Fct_Parameter_At_Pos(string inputStr, int pos)
        {
            objectInfo temp = new objectInfo(false);
            for (int i = 0; i <= CalcFunctions.Count() - 1; ++i)
            {
                objectInfo tempInfo = Is_String_At_Pos(inputStr, CalcFunctions.Get_Function_Parameter(i), pos);
                if (tempInfo.found && tempInfo.len > temp.len)
                    temp = new objectInfo(true, tempInfo.pos, tempInfo.len, inputStr.Substring(tempInfo.pos, tempInfo.len), State.Function);
            }
            return temp;
        }

        static public objectInfo Is_Variable_At_Pos(string inputStr, int pos)
        {
            objectInfo temp = new objectInfo(false);
            for (int i = 0; i <= CalcVariables.Count() - 1; ++i)
            {
                objectInfo tempInfo = Is_String_At_Pos(inputStr, CalcVariables.Get_Var_Name(i), pos);
                if (tempInfo.found && tempInfo.len > temp.len)
                {
                    temp = new objectInfo(true, tempInfo.pos, tempInfo.len, inputStr.Substring(tempInfo.pos, tempInfo.len), State.Variable);
                }
            }
            return temp;
        }

        #endregion
    }
}
