﻿/*
 * ##############################################
 * Script
 * Main part of the QuickCalculator
 * ##############################################
 * 
 * You may copy and distribute this code and the 
 * whole program as you want.
 * This code is not meant to be sold.
 * 
 * Written by Philipp Erler
 * Contact me for feedback, new version or other causes:
 * ph.erler@gmx.net
 * 
 * Version info:
 * Last change: 2011 | 11 | 05
 * .net-Framework: V. 4.0
 * 
 */

using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace CalculationE
{
    static class CalcScript
    {
        /// <summary>
        /// Delegate to call related functions
        /// Script functions will not receive a return value.
        /// Note: Script functions will only submit String-values to to functions.
        /// </summary>
        public delegate void FctDelegate(string[] args);
        //first try; does not work with a function like void test()
        //public delegate void FctDelegate(params object[] obj);

        /// <summary>
        /// Struct to store all script functions' information
        /// </summary>
        struct ScriptFunction
        {
            public string Name;
            public FctDelegate fct;

            public ScriptFunction(string name, FctDelegate fct)
            {
                this.Name = name;
                this.fct = fct;
            }
        }

        /// <summary>
        /// List of all script functions
        /// </summary>
        static List<ScriptFunction> scriptFcts = new List<ScriptFunction>();

        /// <summary>
        /// Add new script function
        /// </summary>
        /// <param name="name"></param>
        /// <param name="fct"></param>
        static public void Add_Fct(string name, FctDelegate fct)
        {
            scriptFcts.Add(new ScriptFunction(name, fct));
        }

        /// <summary>
        /// Check if given string can be a script call.
        /// Check only for script starter character. Allow white space at beginning.
        /// </summary>
        /// <param name="inStr"></param>
        /// <returns></returns>
        static public bool Is_Script_Fct(string inStr)
        {
            string tempStr = inStr;
            tempStr = tempStr.Trim();
            return tempStr.StartsWith(CalcGlobals.SCRIPT_STARTER);
        }

        /// <summary>
        /// Return input string without leading white space and without script starter character
        /// </summary>
        /// <param name="inStr"></param>
        /// <returns></returns>
        static string Process_Input_Str(string inStr)
        {
            string tempStr = inStr;
            tempStr = tempStr.Trim();
            if (!tempStr.StartsWith(CalcGlobals.SCRIPT_STARTER))
                return tempStr;
            tempStr = tempStr.Substring(1).Trim();
            return tempStr;
        }

        /// <summary>
        /// Return index of script function that is called in the given string.
        /// Return -1 if no valid function is found.
        /// </summary>
        /// <param name="inStr">A processed input string.</param>
        /// <returns></returns>
        static int Get_Fct_List_Index(string inStr)
        {
            string tempStr = inStr;

            /*
            //check if script command
            tempStr = tempStr.Trim();
            if (!tempStr.StartsWith(SCRIPT_STARTER.ToString()))
                return -1;
            tempStr = tempStr.Substring(1).Trim();
            */

            //check for existing function
            for (int n = 0; n < scriptFcts.Count; ++n)
            {
                if (tempStr.StartsWith(scriptFcts[n].Name, StringComparison.CurrentCultureIgnoreCase))
                    return n;
            }
            return -1;
        }

        /// <summary>
        /// Execute a script line.
        /// </summary>
        /// <param name="inStr"></param>
        /// <returns>Returns true if script was successfully executed.
        /// It returns false if an error occurred.</returns>
        static public bool Execute_Script(string inStr)
        {
            string tempStr = Process_Input_Str(inStr);
            int fctIndex = Get_Fct_List_Index(tempStr);

            if (fctIndex == -1)
                return false;

            //Remove function name from string
            tempStr = tempStr.Substring(scriptFcts[fctIndex].Name.Length).Trim();
            char[] separators = { ',' };

            //Split remaining string and trim 
            string[] parameters = tempStr.Split(separators, StringSplitOptions.RemoveEmptyEntries);
            for (int n = 0; n < parameters.Count(); ++n)
                parameters[n] = parameters[n].Trim();

            try
            {
                //call delegate
                scriptFcts[fctIndex].fct(parameters);

                return true;
            }
            catch (Exception e)
            {
                throw new Exception(e.Message);
                //return false;
            }
        }

        static public string Concat_String_Array(string[] strArr, string connector)
        {
            string res = "";

            if (strArr == null)
                return res;

            for (int i = 0; i <= strArr.GetUpperBound(0); ++i)
            {
                res += strArr[i];
                if (i < strArr.GetUpperBound(0))
                    res += connector;
            }
            return res;
        }
    }
}
