﻿/*
 * ##############################################
 * Function Drawing
 * Part of the QuickCalculator
 * ##############################################
 * 
 * You may copy and distribute this code and the 
 * whole program as you want.
 * This code is not meant to be sold.
 * 
 * Written by Philipp Erler
 * Contact me for feedback, new version or other causes:
 * ph.erler@gmx.net
 * 
 * Version info:
 * Last change: 2011 | 12 | 28
 * .net-Framework: V. 4.0
 * 
 */


using System;
using System.Drawing;
using System.Drawing.Text;
using System.Windows.Forms;
using System.Collections.Generic;
using CalculationE;

namespace FunctionDrawing
{
    public enum MouseState {Nothing, ZoomBox, MoveOrigin};

    public class FunctionDraw
    {
        //scale and origin offset
        private int originX = 0;                                                    //origin move x
        private int originY = 0;                                                    //origin move y
        private double scaleX = 20;                                                 //scale x; 20 means -10 to +10 fits onto the screen
        private double scaleY = 20;
        private int drawAccuracy = 1;                                               //every x-th pixel is beeing calculated; 5 means 1st pixel is calculated, 6th, 11th, ...

        //inverse y axis; -1 for math default (Ist quadrant); +1 for it default (4th quadrant)
        private int inverseFactor = -1;

        private Rectangle drawScreen;                                               //draw everything within this rect

        private Pen scalePen = new Pen(System.Drawing.Color.Black);                 //black pen for axis and scale numbers
        private List<Pen> functionPens = new List<Pen>();                           //pens for function graphs
        private Pen zoomBoxPen = new Pen(System.Drawing.Color.Gray);                //gray pen for zoombox rect and moveorigin arrow
        private Pen rasterPen = new Pen(System.Drawing.Color.LightGray);            //pen for raster
        private Font scaleFont;                                                     //font for scale numbers; like form.font in italic
        private Font fctFont = new Font("Courier New", 8.0f, FontStyle.Regular);    //font for function names and equation

        //private SolidBrush fctBrush = new SolidBrush(System.Drawing.Color.Black);   //black solid brush for function text
        private SolidBrush scaleBrush = new SolidBrush(System.Drawing.Color.Black); //black solid brush for scale numbers
        private SolidBrush arrowBrush = new SolidBrush(System.Drawing.Color.Gray);  //gray solid brush for move arrow

        private Rectangle zoomBox = new Rectangle(-1, -1, 0, 0);                    //rect for zoombox and moveorigin; -1 to mark as empty
        private MouseState mouseState = MouseState.Nothing;                         //left or right mouse button

        //Note: The function shown on this form do NOT neccessarily have the same indiezes as the functions on CalcFunctions
        public List<string> fctNames;                                               //name of function
        public List<string> fctPrint;                                               //complete description of each function
        public List<int> fctIndiezes;                                               //Indiezes of corresponding function in calcfunctions' list
        public List<bool> fctSelect;                                                //Selected state of user functions

        private Point[][] functionPoints;                                           //jagged array to save calculation results
        private bool validFctGraph = true;

        //progress bar info
        private ToolStripProgressBar progressBar = null;                            //progress bar "pointer"
        private int calcProgressVal = 0;
        private int calcProgressMin = 0;
        private int calcProgressMax = 1;

        public FunctionDraw(int width, int height)
        {
            drawScreen = new Rectangle(0, 0, width, height);                        //draw screen gets size of picturebox
            scaleFont = new Font("Courier New", 9.0f, FontStyle.Italic);            //set scale font to form.font
        }

        public void Actualize_Functions()
        {
            fctNames = CalcFunctions.Get_User_Function_Names();
            fctPrint = CalcFunctions.Get_User_Functions();
            fctIndiezes = CalcFunctions.Get_User_Function_Indiezes();
            fctSelect = new List<bool>(fctIndiezes.Count);
            for (int i = 0; i < fctIndiezes.Count; ++i)
                fctSelect.Add(false);
        }

        #region Get_Set_Vars
        public MouseState Get_Mouse_State()
        {   return this.mouseState;     }

        public void Set_Mouse_State(MouseState state)
        {   this.mouseState = state;    }

        public void Set_Progress_Bar(ToolStripProgressBar pBar)
        {   progressBar = pBar;         }

        public int Get_Progress_Min()
        {   return calcProgressMin;     }

        public int Get_Progress_Max()
        {   return calcProgressMax;     }

        public int Get_Progress_Val()
        {   return calcProgressVal;     }

        public double Get_Scale_X()
        {   return this.scaleX;         }

        public double Get_Scale_Y()
        {   return this.scaleY;         }

        public void Set_Scale_X(double val)
        {
            this.scaleX = val;
            this.validFctGraph = false;
        }

        public void Set_Scale_Y(double val)
        {
            this.scaleY = val;
            this.validFctGraph = false;
        }

        public int Get_Origin_X()
        {   return this.originX;        }

        public int Get_Origin_Y()
        {   return this.originY;        }

        public void Set_Origin_X(int val)
        {
            this.originX = val;
            this.validFctGraph = false;
        }

        public void Set_Origin_Y(int val)
        {
            this.originY = val;
            this.validFctGraph = false;
        }

        public int Get_Accurracy()
        {
            return this.drawAccuracy;
        }

        public void Set_Accurracy(int val)
        {
            this.drawAccuracy = val;
            this.validFctGraph = false;
        }

        public int Get_Inverse_Factor()
        {
            return this.inverseFactor;
        }

        public void Set_Inverse_Factor(int val)
        {
            this.inverseFactor = val;
            this.validFctGraph = false;
        }

        public int Get_Zoom_Box_X()
        {
            return this.zoomBox.X;
        }

        public void Set_Zoom_Box_X(int val)
        {
            this.zoomBox.X = val;
        }

        public int Get_Zoom_Box_Y()
        {
            return this.zoomBox.Y;
        }

        public void Set_Zoom_Box_Y(int val)
        {
            this.zoomBox.Y = val;
        }

        public int Get_Zoom_Box_Width()
        {
            return this.zoomBox.Width;
        }

        public void Set_Zoom_Box_Width(int val)
        {
            this.zoomBox.Width = val;
        }

        public int Get_Zoom_Box_Height()
        {
            return this.zoomBox.Height;
        }

        public void Set_Zoom_Box_Height(int val)
        {
            this.zoomBox.Height = val;
        }

        public int Get_Zoom_Box_Right()
        {
            return this.zoomBox.Right;
        }

        public int Get_Zoom_Box_Bottom()
        {
            return this.zoomBox.Bottom;
        }
        #endregion

        //draw axis and scale numbers
        public void Draw_Scale(Graphics Graphic)
        {
            double scaleXStep = 0;
            double scaleYStep = 0;

            scaleXStep = (scaleX / 10);
            scaleYStep = (scaleY / 10);
            for (int i = -20; i <= 20; ++i)
            {
                try
                {

                    int nextMarkPosX = (int)(((i * scaleXStep * drawScreen.Width) / scaleX) + 0.5 * drawScreen.Width + originX);

                    int nextMarkPosY = (int)(((i * scaleYStep * drawScreen.Height) / scaleY) + 0.5 * drawScreen.Height + originY * inverseFactor);

                    //y-direction raster lines
                    Graphic.DrawLine(rasterPen, nextMarkPosX, 0, nextMarkPosX, drawScreen.Bottom);

                    //x-direction raster lines
                    Graphic.DrawLine(rasterPen, 0, nextMarkPosY, drawScreen.Right, nextMarkPosY);

                    //x axis
                    Graphic.DrawLine(scalePen, drawScreen.Left, (int)(0.5 * drawScreen.Bottom + inverseFactor * originY),
                                                  drawScreen.Right, (int)(0.5 * drawScreen.Bottom + inverseFactor * originY));
                    //y axis
                    Graphic.DrawLine(scalePen, (int)(0.5 * drawScreen.Right + originX), drawScreen.Top,
                                                  (int)(0.5 * drawScreen.Right + originX), drawScreen.Bottom);

                    //x-axis scale marks
                    Graphic.DrawLine(scalePen, nextMarkPosX, (int)(inverseFactor * -drawScreen.Height * 0.02 + inverseFactor * originY + 0.5 * drawScreen.Width),
                                                  nextMarkPosX, (int)(inverseFactor * drawScreen.Height * 0.02 + inverseFactor * originY + 0.5 * drawScreen.Width));

                    //y-axis scale marks
                    Graphic.DrawLine(scalePen, (int)(0.5 * drawScreen.Height + originX - drawScreen.Width * 0.02), nextMarkPosY,
                                                  (int)(0.5 * drawScreen.Height + originX + drawScreen.Width * 0.02), nextMarkPosY);
                    //x-axis scale numbers
                    Graphic.DrawString((i * scaleXStep).ToString(), scaleFont, scaleBrush, nextMarkPosX,
                                            (int)(-10 + -drawScreen.Height * 0.02 + 0.5 * drawScreen.Width + inverseFactor * originY));
                    //y-axis scale numbers
                    Graphic.DrawString((i * scaleYStep * inverseFactor).ToString(), scaleFont, scaleBrush,
                                            (int)(0.5 * drawScreen.Height + originX - 10 - drawScreen.Width * 0.02), nextMarkPosY);

                }
                catch (Exception)
                { return; }
            }
        }

        //draw functions
        public void Draw_Functions(Graphics Graphic)
        {
            if (functionPoints == null || validFctGraph == false)
                return;

            //all functions
            for (int i = 0; i <= functionPoints.GetUpperBound(0); ++i)
            {
                if (fctSelect[i])   //draw only if selected
                {
                    //all points
                    for (int n = 0; n < functionPoints[i].GetUpperBound(0); ++n)
                    {
                        if (functionPoints[i][n].IsEmpty == false && functionPoints[i][n + 1].IsEmpty == false)
                        {
                            try
                            {
                                Graphic.DrawLine(functionPens[i], functionPoints[i][n], functionPoints[i][n + 1]);
                            }
                            catch (Exception)
                            {
                                //might be overflow
                            }
                        }
                    }
                }
            }
        }

        public void Draw_Function_Texts(Graphics g)
        {
            float heightOfLine = fctFont.GetHeight(g) + 3.0f;
            float numLines = 0.0f;
            SolidBrush fctBrush;

            for (int i = 0; i < fctPrint.Count; ++i)
            {
                if (fctSelect[i])   //draw only if selected
                {
                    fctBrush = new SolidBrush(functionPens[i].Color);
                    g.DrawString(CalcFunctions.Print_Function(fctIndiezes[i]), fctFont, fctBrush, new PointF(5.0f, heightOfLine * numLines));
                    numLines += 1.0f;
                }
            }
        }

        public void Set_Fct_Pens(List<Pen> fctPenList)
        {
            functionPens = fctPenList;
        }

        public string Get_Mouse_Coord_X(int x)
        {
            //get x coordinate
            double tempDou = (x - originX) / (drawScreen.Width / scaleX) - scaleX / 2;
            //format and show in status bar

            //such a shit
            //normal format function isn't working here because of round values
            string tempD = tempDou.ToString("0000.0000");
            string signed = tempDou.ToString();
            if (signed[0] != '-')
                tempD = "+" + tempD;
            else if (signed[0] == '-' && tempD[0] != '-')
                tempD = "-" + tempD;

            return tempD;
        }

        public string Get_Mouse_Coord_Y(int y)
        {
            //get y coordinate
            double tempDou = (inverseFactor * y - originY) / (drawScreen.Height / scaleY) - inverseFactor * scaleY / 2;
            //format and show in status bar
            string tempD = tempDou.ToString("0000.0000");
            string signed = tempDou.ToString();
            if (signed[0] != '-')
                tempD = "+" + tempD;
            else if (signed[0] == '-' && tempD[0] != '-')
                tempD = "-" + tempD;

            return tempD;
        }

        public void Actualize_Zoom_Box(int x, int y)
        {
            //actualize zoomBox size
            if (zoomBox.X > -1)
            {
                zoomBox.Width = x - zoomBox.X;
                zoomBox.Height = y - zoomBox.Y;
            }
        }

        //calculate points of all functions
        public void Calc_Functions()
        {
            bool writeLog = CalcEngine.Get_Write_Log();
            int NumPoints = 1 + (int)(drawScreen.Width / drawAccuracy);     //number of points on every function graph
            int currPIndex = 0;                                             //index of current point
            functionPoints = new Point[fctIndiezes.Count][];                //set 1st dimension
            Point currentPoint = new Point(0, 0);                           //current point
            double result = 0.0;                                            //result of calculation
            double setInValue = 0;                                          //value to set in the function; convert pixel to this value

            if (fctPrint.Count == 0)
                return;

            CalcEngine.Set_Write_Log(false);

            //for functions
            for (int fctIndex = 0; fctIndex < fctIndiezes.Count; ++fctIndex)
            {
                if (fctSelect[fctIndex])   //calc only if selected
                {
                    functionPoints[fctIndex] = new Point[NumPoints];            //set 2nd dimension to number of points

                    //progress bar starting point
                    if (progressBar != null)
                    {
                        progressBar.Minimum = 0;
                        progressBar.Maximum = NumPoints;
                        progressBar.Value = 0;
                    }
                    calcProgressMin = 0;
                    calcProgressMax = NumPoints;
                    calcProgressVal = 0;

                    //for points
                    currPIndex = 0;
                    for (int x = 0; x < drawScreen.Width + drawAccuracy - 1; x += drawAccuracy)
                    {
                        try
                        {
                            //convert pixel in scale units; save in setinvalue
                            setInValue = ((x - originX - 0.5 * drawScreen.Width) / (drawScreen.Width / scaleX));
                            //perform calculation   nestedlevel should not matter here
                            result = CalcFunctions.Calc_Function(new CalculationElement(
                                CalculationElement.ElementType.Function, fctIndiezes[fctIndex], -1), setInValue).GetValue();
                            //set current point
                            currentPoint.X = x;
                            currentPoint.Y = (int)Math.Round(inverseFactor * result * (drawScreen.Height / scaleY) +
                                                          inverseFactor * originY + 0.5 * drawScreen.Height, 0);

                            //save point in array;
                            functionPoints[fctIndex][currPIndex] = currentPoint;

                            //progress bar
                            calcProgressVal++;
                        }
                        catch (Exception)   //calculation may return an error message
                        {                   //set current point to empty; this point will be ignored when drawing the graph
                            functionPoints[fctIndex][currPIndex] = new Point();
                        }
                        currPIndex++;

                        if (progressBar != null)
                            progressBar.PerformStep();
                    }
                    if (progressBar != null)
                        progressBar.Value = 0;
                    calcProgressVal = 0;                        //reset progress bar
                }
            }
            validFctGraph = true;                       //graph is now valid
            CalcEngine.Set_Write_Log(writeLog);
        }

        public void Draw_All(Graphics graph)
        {
            graph.Clear(System.Drawing.Color.White);            //clear 
            this.Draw_Scale(graph);                             //draw scale and numbers
            this.Draw_Functions(graph);                         //draw function graphs
            this.Draw_Function_Texts(graph);                    //draw texts

            //gdi cannot draw rects with negative height or width; so change values in valid ones
            Rectangle tempRect = zoomBox;                       //contains changed values
            if (this.mouseState == MouseState.ZoomBox)          //zoomBox
            {   //switch left<->right; top<->bottom if neccessary
                if (zoomBox.X > -1)                             //skip if empty
                {
                    if (tempRect.Width < 0)
                    {
                        tempRect.X += tempRect.Width;
                        tempRect.Width = Math.Abs(tempRect.Width);
                    }
                    if (tempRect.Height < 0)
                    {
                        tempRect.Y += tempRect.Height;
                        tempRect.Height = Math.Abs(tempRect.Height);
                    }
                    graph.DrawRectangle(zoomBoxPen, tempRect); //draw
                }
            }
            else if (this.mouseState == MouseState.MoveOrigin)  //MoveArrow
            {
                if (zoomBox.X > -1)                             //skip if empty
                {
                    //link line
                    graph.DrawLine(scalePen, zoomBox.Left, zoomBox.Top, zoomBox.Right, zoomBox.Bottom);
                    //arrow head
                    Point[] ArrowHead = new Point[3];
                    ArrowHead[0] = new Point(zoomBox.Right, zoomBox.Bottom);

                    double angleArrow = Math.Atan2(zoomBox.Height, zoomBox.Width);
                    double angleHead = angleArrow - (Math.PI / 2);

                    ArrowHead[1] = new Point((int)(zoomBox.X + zoomBox.Width * 0.9 + Math.Cos(angleHead) * 10),
                                            (int)(zoomBox.Y + zoomBox.Height * 0.9 + Math.Sin(angleHead) * 10));

                    ArrowHead[2] = new Point((int)(zoomBox.X + zoomBox.Width * 0.9 - Math.Cos(angleHead) * 10),
                                            (int)(zoomBox.Y + zoomBox.Height * 0.9 - Math.Sin(angleHead) * 10));

                    graph.FillPolygon(arrowBrush, ArrowHead);   //draw
                }
            }
        }

        public void Resolve_Zoom_Box()
        {
            //get new scale
            this.scaleX = Math.Abs(this.scaleX * this.zoomBox.Width / drawScreen.Width);
            this.scaleY = Math.Abs(this.scaleY * this.zoomBox.Height / drawScreen.Height);

            //get new origin
            this.originX = this.originX - (int)Math.Round((this.zoomBox.X + 0.5 * this.zoomBox.Width - drawScreen.Width * 0.5) * drawScreen.Width / this.zoomBox.Width);
            this.originY = this.originY - (int)Math.Round((this.zoomBox.Y + 0.5 * this.zoomBox.Height - drawScreen.Height * 0.5) * this.inverseFactor * drawScreen.Height / this.zoomBox.Height);

            //mark as empty
            this.mouseState = MouseState.Nothing;
        }

        public void Resolve_Move_Origin()
        {
            //get new origin
            this.originX = this.originX - this.zoomBox.Right + this.zoomBox.X;
            this.originY = this.originY + this.inverseFactor * (-this.zoomBox.Bottom + this.zoomBox.Y);

            //mark as empty
            this.mouseState = MouseState.Nothing;
        }

    }
}