﻿/*
 * ##############################################
 * MainForm-Form
 * Part of the QuickCalculator
 * ##############################################
 * 
 * You may copy and distribute this code and the 
 * whole program as you want.
 * This code is not meant to be sold.
 * 
 * Written by Philipp Erler
 * Contact me for feedback, new version or other causes:
 * ph.erler@gmx.net
 * 
 * Version info:
 * Last change: 2013 | 02 | 03
 * .net-Framework: V. 4.0
 * 
 * You can compile this for Mono on Mac OS and Linux using MonoDevelop.
 * You must switch in MonoDevelop the target Framework to "Mono / .Net 4.0".
 * Linux and Mac do not support icons on forms.
 * See http://superuser.com/questions/67126/how-to-associate-the-exe-extension-to-be-opened-with-mono
 * to know how to associate .exe with mono on Macs.
 * 
 * TODO:
 * Bugs:
 *      --#draw does not refresh list of functions in draw window
 *      --allow other brackets [{]} ...
 *      **f5 on draw window does not refresh function list
 *      --n. Def.:  -e ^ (1/5)  = -1.2214
 *      --0 | x returns now Error: Division by zero
 *      --clear message when calculation returns a complex result
 *      
 * cotangent
 * 
 * Fullscreen for drawfunction
 * drawfunctions selection changed: recalc only changed functions
 * remove jump at poles
 * function draw: zoom also by mouse wheel and touchpad gesture
 * 
 * additional script commands:
 *      drawcolor function colorcode
 *      delvar varname
 *      delfct fctname
 *      delvars
 *      delfcts
 *      
 * Nice to have:
 *      Equation solver
 *      3 dimensional function drawing
 *          max(x, y) = (x + y + abs(x - y)) / 2
 *      (graphical) differentiation
 *      (graphical) integration
 *      Link .qcalc to qcalc (registry) ; dll for admin rights; extras-menu
 *      Remake intellikeyboard
 *      user defined integer base conversion; e.g. 20 base dec -> 1 base 20
 */

using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Windows.Forms;
using System.IO;
using CalculationE;

namespace QuickCalculator
{
    public partial class MainForm : Form
    {
        //draw form
        static public frmDrawFunction frmDrawFct;

        //about form
        static public Form frmAbout;

        //arrays for menuitems to delete var and function
        private List<ToolStripMenuItem> delVarMenuItems = new List<ToolStripMenuItem>();
        private List<ToolStripMenuItem> delFctMenuItems = new List<ToolStripMenuItem>();

        private UInt32 numEntries = 0;

        //constructor
        public MainForm()
        {
            Init_MainForm();
        }

        //constructor for command-line args
        public MainForm(string[] args)
        {
            Init_MainForm();

            for (int i = 0; i < args.Count<string>();  ++i)
            {
                if (System.IO.File.Exists(args[i]))
                    Load_File(args[i]);
                else
                    Execute_Line(args[i]);
            }
        }

        /// <summary>
        /// Own initialization function for all constructors.
        /// </summary>
        private void Init_MainForm()
        {
            Initialize_MainForm();
            SetScriptFunctions();
            Execute_Startup_Files();
            Switch_To_Singleline();
            MainForm_Resize(this, new EventArgs());
        }

        ~MainForm()
        { }

        /// <summary>
        /// If existing execute "init.qcalc" from currentuser/documents/qcalc/
        /// Else execute fallback file "init.qcalc" from the application's folder
        /// </summary>
        private void Execute_Startup_Files()
        {
            string path = Environment.GetFolderPath(Environment.SpecialFolder.MyDocuments) + @"/QCalc/init.qcalc";
            if (File.Exists(path))
            {
                Load_File(path);
                Execute_List(Table_To_List());
                rtbInput.Text = "";
            }
            else
            {
                path = Application.StartupPath + @"/init.qcalc";
                if (File.Exists(path))
                {
                    Load_File(path);
                    Execute_List(Table_To_List());
                    rtbInput.Text = "";
                }
            }
        }

        private void SetScriptFunctions()
        {
            CalcScript.Add_Fct("clr", new CalcScript.FctDelegate(clearList));
            CalcScript.Add_Fct("draw", new CalcScript.FctDelegate(Open_Window_Draw_Fcts));
            CalcScript.Add_Fct("log", new CalcScript.FctDelegate(Make_Log_Entry));
            CalcScript.Add_Fct("echo", new CalcScript.FctDelegate(Write_ListBox_Entry));
            CalcScript.Add_Fct("setorigin", new CalcScript.FctDelegate(Set_Draw_Origin));
            CalcScript.Add_Fct("setscale", new CalcScript.FctDelegate(Set_Draw_Scale));
            CalcScript.Add_Fct("redraw", new CalcScript.FctDelegate(Fct_Redraw));
            CalcScript.Add_Fct("exec", new CalcScript.FctDelegate(Execute_File));
            CalcScript.Add_Fct("decsign", new CalcScript.FctDelegate(Set_Decimal_Sign));
            CalcScript.Add_Fct("anglemode", new CalcScript.FctDelegate(Set_Angle_Mode));
            CalcScript.Add_Fct("multiline", new CalcScript.FctDelegate(Set_Multiline_State));
            CalcScript.Add_Fct("outformat", new CalcScript.FctDelegate(Set_Output_Format));
            CalcScript.Add_Fct("exit", new CalcScript.FctDelegate(Exit_Application));
        }

        private void Initialize_MainForm()
        {
            InitializeComponent();
            toolStripStatusLabelVersion.Text = CalculationE.CalcGlobals.QCALCVERSION;
            writeLogToolStripMenuItem.PerformClick();
            Actualize_DecSign();
            rtbInput.Focus();
        }

        private void Actualize_DecSign()
        {
            //decimal sign
            if (CalcEngine.Get_Decimal_Sign().NumberFormat.NumberDecimalSeparator == ",")
            {
                commaToolStripMenuItem.Checked = true;
                pointToolStripMenuItem.Checked = false;
            }
            else
            {
                commaToolStripMenuItem.Checked = false;
                pointToolStripMenuItem.Checked = true;
            }
        }

        public static bool Is_Form_Open(string formName)
        {
            foreach (Form OpenForm in Application.OpenForms)
            {
                if (OpenForm.Name == formName)
                    return true;
            }

            return false;
        }

        #region Script_Function_Targets

        //no exception catches here because the script system should notice them

        private void Exit_Application(string[] args)
        { Application.Exit(); }

        //script function
        //dummy function; output is generated in exec script function
        private void Write_ListBox_Entry(string[] input)
        {/*
            string entry;

            entry = CalcScript.Concat_String_Array(input, ", ");

            //print input string
            Add_Calc_List_Entry(entry);
          */
        }

        //script function
        private void Make_Log_Entry(string[] args)
        {
            CalcEngine.Add_Line_Log(CalcScript.Concat_String_Array(args, ", "));
            CalcEngine.Print_Log(Application.StartupPath + @"/log.txt");
        }

        private void clearList(string[] str)
        {
            dgvRes.Rows.Clear();
            numEntries = 0;
            for (int n = 0; n < dgvRes.Rows.Count; ++n)
            {
                dgvRes.Rows[n].Cells[1].Style.Alignment = DataGridViewContentAlignment.MiddleLeft;
                dgvRes.Rows[n].Cells[1].Style.ForeColor = Color.Black;
            }
            Reset_Status_Label();
        }

        private void Open_Window_Draw_Fcts(string[] args)
        {
            Open_DrawForm();
            frmDrawFct.Select_Fcts(args);
        }

        private void Set_Draw_Origin(string[] args)
        {
            frmDrawFct.Set_Origin(Convert.ToInt32(args[0], CalcEngine.Get_Decimal_Sign().NumberFormat),
                                  Convert.ToInt32(args[1], CalcEngine.Get_Decimal_Sign().NumberFormat));
        }

        private void Set_Draw_Scale(string[] args)
        {
            frmDrawFct.Set_Scale(Convert.ToDouble(args[0], CalcEngine.Get_Decimal_Sign().NumberFormat),
                                 Convert.ToDouble(args[1], CalcEngine.Get_Decimal_Sign().NumberFormat));
        }

        private void Fct_Redraw(string[] args)
        {
            frmDrawFct.Fct_Redraw();
        }

        private void Execute_File(string[] args)
        {
            bool saveMultiline = multilineToolStripMenuItem.Checked;
            rtbInput.Text = "";

            Execute_List(File_To_String_Array(CalcScript.Concat_String_Array(args, ", ")));

            if (saveMultiline)
                Switch_To_Multiline();
            else
                Switch_To_Singleline();
            rtbInput.Text = "";
        }

        private void Set_Decimal_Sign(string[] args)
        {
            if (args[0].StartsWith("comma", StringComparison.CurrentCultureIgnoreCase))
                Set_DecSign_Comma();
            else if (args[0].StartsWith("point", StringComparison.CurrentCultureIgnoreCase))
                Set_DecSign_Point();
            else if (args[0].StartsWith("default", StringComparison.CurrentCultureIgnoreCase))
                if (System.Globalization.CultureInfo.CurrentCulture.NumberFormat.NumberDecimalSeparator == ",")
                    Set_DecSign_Comma();
                else
                    Set_DecSign_Point();
            else
                throw new Exception("Invalid decimal sign.");
        }

        private void Set_Angle_Mode(string[] args)
        {
            if (args[0].StartsWith("rad", StringComparison.CurrentCultureIgnoreCase))
                Set_Angle_Mode_Radiant();
            else if (args[0].StartsWith("deg", StringComparison.CurrentCultureIgnoreCase))
                Set_Angle_Mode_Degree();
            else
                throw new Exception("Invalid angle mode.");
        }

        private void Set_Multiline_State(string[] args)
        {
            if (args[0].StartsWith("multi", StringComparison.CurrentCultureIgnoreCase))
                Switch_To_Multiline();
            else if (args[0].StartsWith("single", StringComparison.CurrentCultureIgnoreCase))
                Switch_To_Singleline();
            else
                throw new Exception("Invalid multiline mode.");
        }

        private void Set_Output_Format(string[] args)
        {
            if (args[0].StartsWith("dec", StringComparison.CurrentCultureIgnoreCase))
                setOutputFormat(CalcEngine.OutputFormat.Decimal);
            else if (args[0].StartsWith("exp", StringComparison.CurrentCultureIgnoreCase))
                setOutputFormat(CalcEngine.OutputFormat.Exponential);
            else if (args[0].StartsWith("hex", StringComparison.CurrentCultureIgnoreCase))
                setOutputFormat(CalcEngine.OutputFormat.Hexadecimal);
            else if (args[0].StartsWith("oct", StringComparison.CurrentCultureIgnoreCase))
                setOutputFormat(CalcEngine.OutputFormat.Octal);
            else if (args[0].StartsWith("bin", StringComparison.CurrentCultureIgnoreCase))
                setOutputFormat(CalcEngine.OutputFormat.Binary);
            else
                throw new Exception("Invalid output format.");
        }

        #endregion

        #region Calculation_List

        //Add line to calc list
        private void Add_Calc_List_Entry(string entry)
        {
            if (entry == "")
                return;

            // limit number of entries
            if (dgvRes.Rows.Count > 300)
                dgvRes.Rows.RemoveAt(0);

            // add new entry
            dgvRes.Rows.Add();
            ++numEntries;
            dgvRes.Rows[dgvRes.Rows.Count - 1].Cells[0].Value = numEntries.ToString();
            dgvRes.Rows[dgvRes.Rows.Count - 1].Cells[1].Value = entry;
            dgvRes.FirstDisplayedScrollingRowIndex = dgvRes.Rows.Count - 1;
        }

        private void dgvRes_CellDoubleClick(object sender, DataGridViewCellEventArgs e)
        {
            String copyString = "";

            copyString = dgvRes.SelectedRows[0].Cells[1].Value.ToString();   //get entry

            if (multilineToolStripMenuItem.Checked)
            {
                Input_Str_Insertion(copyString);
            }
            else
            {
                dgvRes.ClearSelection();
                rtbInput.Text = copyString;
                rtbInput.Focus();										            //set focus on rtbInput
                rtbInput.SelectionStart = rtbInput.Text.Length;
            }
        }

        #endregion

        //insert string into rtbInput.text at selection
        private void Input_Str_Insertion(string input)
        {
            if (input == "")
                return;

            int saveSelStart = rtbInput.SelectionStart;

            rtbInput.Text = rtbInput.Text.Remove(rtbInput.SelectionStart, rtbInput.SelectionLength);
            rtbInput.Text = rtbInput.Text.Insert(saveSelStart, input);
            rtbInput.SelectionStart = saveSelStart + input.Length;
        }

        //start calculation
        private void calculateToolStripMenuItem_Click(object sender, EventArgs e)
        {
            Start_Calculation();
        }

        private void rtbInput_KeyDown(object sender, KeyEventArgs e)
        {
            if (!multilineToolStripMenuItem.Checked && e.KeyCode == Keys.Enter)
            {
                Start_Calculation();
                e.Handled = true;
            }
        }

        private void SetOutputFormat()
        {
            if (exponentialViewToolStripMenuItem.Checked)
                CalcEngine.Set_Output_Format(CalcEngine.OutputFormat.Exponential);
            else if (decimalToolStripMenuItem.Checked)
                CalcEngine.Set_Output_Format(CalcEngine.OutputFormat.Decimal);
            else if (binaryToolStripMenuItem.Checked)
                CalcEngine.Set_Output_Format(CalcEngine.OutputFormat.Binary);
            else if (octalToolStripMenuItem.Checked)
                CalcEngine.Set_Output_Format(CalcEngine.OutputFormat.Octal);
            else if (hexadecimalToolStripMenuItem.Checked)
                CalcEngine.Set_Output_Format(CalcEngine.OutputFormat.Hexadecimal);
        }

        private void Print_CalcResults(List<OutputElement> results)
        {
            foreach (OutputElement oEl in results)
            {
                if (oEl.GetOutputType() == OutputElement.OutputType.CalcInput)
                {
                    Add_Calc_List_Entry(oEl.GetMsg());
                }
                else if (oEl.GetOutputType() == OutputElement.OutputType.CalcOutput)
                {
                    Add_Calc_List_Entry(oEl.GetMsg());
                    dgvRes.Rows[dgvRes.Rows.Count - 1].Cells[1].Style.Alignment = DataGridViewContentAlignment.MiddleRight;
                }
                else if (oEl.GetOutputType() == OutputElement.OutputType.ControlStructure)
                {
                    Add_Calc_List_Entry(oEl.GetMsg());
                }
                else if (oEl.GetOutputType() == OutputElement.OutputType.ScriptInput)
                {
                    //Add_Calc_List_Entry(oEl.GetMsg());
                }
                else if (oEl.GetOutputType() == OutputElement.OutputType.ScriptMSG)
                {
                    Add_Calc_List_Entry(oEl.GetMsg());
                }
                else if (oEl.GetOutputType() == OutputElement.OutputType.Error)
                {
                    Add_Calc_List_Entry(oEl.GetMsg());
                    dgvRes.Rows[dgvRes.Rows.Count - 1].Cells[1].Style.Alignment = DataGridViewContentAlignment.MiddleRight;
                    dgvRes.Rows[dgvRes.Rows.Count - 1].Cells[1].Style.ForeColor = Color.Red;
                }
                else
                {
                }
            }
        }

        private void Start_Calculation()
        {
            DateTime startTime;
            DateTime endTime;

            toolStripStatusLabelState.Text = "Working...";

            if (rtbInput.Text == "")                                            //if empty 
            {                                                                   //do last calculation again
                try
                {
                    if (dgvRes.Rows.Count == 1)
                        rtbInput.Text = dgvRes.Rows[dgvRes.Rows.Count - 1].Cells[1].Value.ToString();
                    else if (dgvRes.Rows.Count > 1)
                        rtbInput.Text = dgvRes.Rows[dgvRes.Rows.Count - 2].Cells[1].Value.ToString();
                }
                catch (Exception)
                {
                    toolStripStatusLabelState.Text = "Ready";
                    return;
                }
            }

            startTime = DateTime.Now;

            Verify_Round_Decimals();
            int decimals = System.Convert.ToInt32(roundToolStripTextBox.Text);
            CalcEngine.Set_Num_Decimals(decimals);
            SetOutputFormat();

            string[] inputs = Table_To_List();
            string savedInput = rtbInput.Text;

            Print_CalcResults(CalcEngine.Exec_Script(inputs));

            //ready for next calculation
            if (multilineToolStripMenuItem.Checked == false)
                rtbInput.Clear();

            endTime = DateTime.Now;
            //Do not show passed time if calc list is empty (=> #clr was executed)
            if (dgvRes.Rows.Count == 0)
                Reset_Status_Label();
            else
            {
                //Show calculation duration
                toolStripStatusLabelState.Text = "Calculation time: " +
                    endTime.Subtract(startTime).TotalMilliseconds.ToString(CalcEngine.Get_Decimal_Sign().NumberFormat) + " ms";
                CalcEngine.Print_Log(Application.StartupPath + @"/log.txt");
            }

            //set focus on rtbInput if frmdraw did not get the focus
            if (frmDrawFct != null)
            {
                if (!frmDrawFct.ContainsFocus)
                    rtbInput.Focus();
            }
            else
                rtbInput.Focus();
        }

        private void saveToFileToolStripMenuItem_Click(object sender, EventArgs e)
        {
            if (saveFileDialog.ShowDialog() == DialogResult.OK)
            {
                try
                {
                    System.IO.File.WriteAllText(saveFileDialog.FileName, rtbInput.Text);
                }
                catch (Exception ex)
                {
                    MessageBox.Show("An Error occurred while saving: " + ex.Message);
                }
            }
        }

        private void Load_File(string path)
        {
            string[] fileContent;

            try
            {
                fileContent = System.IO.File.ReadAllLines(path);
            }
            catch (System.IO.FileNotFoundException)
            {
                MessageBox.Show("File not found: " + path);
                return;
            }
            catch (System.Security.SecurityException)
            {
                MessageBox.Show("You do not have the permission to access the file: " + path);
                return;
            }
            catch (Exception ex)
            {
                MessageBox.Show("An Error occurred while loading file: " + ex.Message);
                return;
            }

            Switch_To_Multiline();

            rtbInput.Text = CalcScript.Concat_String_Array(fileContent, Environment.NewLine);

        }

        private string[] File_To_String_Array(string path)
        {
            string[] fileContent;

            try
            {
                fileContent = System.IO.File.ReadAllLines(path);
                return fileContent;
            }
            catch (System.IO.FileNotFoundException)
            {
                throw new Exception("File not found: " + path);
            }
            catch (System.Security.SecurityException)
            {
                throw new Exception("You do not have the permission to access the file: " + path);
            }
            catch (Exception ex)
            {
                throw new Exception("An Error occurred while loading file: " + ex.Message);
            }
        }

        private void Execute_List(string[] list)
        {
            foreach (string line in list)
            {
                if (line != "")
                {
                    try
                    {
                        Execute_Line(line);
                    }
                    catch (Exception ex)
                    {
                        MessageBox.Show("Error: " + ex.Message);
                        return;
                    }
                }
            }

        }

        //return a list containing all lines
        private string[] Table_To_List()
        {
            string[] separators = { "\n" };
            string[] lines = rtbInput.Text.Split(separators, StringSplitOptions.None);

            return lines;
        }

        private string Print_Rounded_Val(double val)
        {
            Verify_Round_Decimals();
            return Math.Round(val, System.Convert.ToInt32(roundToolStripTextBox.Text)).ToString();
        }

        private void Verify_Round_Decimals()
        {
            bool success;
            int decimals;
            success = int.TryParse(roundToolStripTextBox.Text, out decimals);
            if (!success || decimals < 0 || decimals > 15)
            {
                roundToolStripTextBox.Text = "10";
            }
            //TODO: show error message
        }

        private void MainForm_DragDrop(object sender, DragEventArgs e)
        {
            if (e.Data.GetDataPresent(DataFormats.Text))
            {
                string lines = e.Data.GetData(DataFormats.Text).ToString();
                if (lines.IndexOf(Environment.NewLine) > 0)
                    Switch_To_Multiline();
                else
                    Switch_To_Singleline();

                rtbInput.Text = lines;
            }
            else if (e.Data.GetDataPresent(DataFormats.FileDrop))
            {
                string[] files = (string[])e.Data.GetData(DataFormats.FileDrop);
                foreach (string file in files)
                {
                    if (file != "")
                        Load_File(file);
                }
            }
        }

        #region Multiline

        const float rtbInputBigFontSize = 8.25f;
        const float rtbInputSmallFontSize = 20.0f;

        private void multilineToolStripMenuItem_Click(object sender, EventArgs e)
        {
            if (!multilineToolStripMenuItem.Checked)   // multiline mode
            {
                Switch_To_Multiline();
            }
            else
            {
                Switch_To_Singleline();
            }
        }

        private void Switch_To_Multiline()
        {
            multilineToolStripMenuItem.Checked = true;
            rtbInput.Font = new Font(FontFamily.GenericMonospace, rtbInputBigFontSize);
            rtbInput.Multiline = true;
            rtbInput.ScrollBars = RichTextBoxScrollBars.ForcedBoth;
            MainForm_Resize(this, new EventArgs());
        }

        private void Switch_To_Singleline()
        {
            multilineToolStripMenuItem.Checked = false;
            rtbInput.Font = new Font(FontFamily.GenericMonospace, rtbInputSmallFontSize);
            rtbInput.Multiline = false;
            rtbInput.ScrollBars = RichTextBoxScrollBars.None;
            MainForm_Resize(this, new EventArgs());
        }

        private void MainForm_Resize(object sender, EventArgs e)
        {
            const int rtbInputSmallHeight = 38;
            const int padding = 8;
            
            if (multilineToolStripMenuItem.Checked)
            {
                rtbInput.Left = padding;
                rtbInput.Top = menuStrip.Size.Height + padding;
                rtbInput.Width = menuStrip.Width - 2 * padding;
                rtbInput.Height = (statusStrip.Top - menuStrip.Size.Height - 3 * padding) / 2;

                dgvRes.Left = padding;
                dgvRes.Top = rtbInput.Bottom + padding;
                dgvRes.Width = menuStrip.Width - 2 * padding;
                dgvRes.Height = (statusStrip.Top - menuStrip.Size.Height - 3 * padding) / 2;

            }
            else
            {
                rtbInput.Left = padding;
                rtbInput.Top = menuStrip.Size.Height + padding;
                rtbInput.Width = menuStrip.Width - 2 * padding;
                rtbInput.Height = rtbInputSmallHeight;

                dgvRes.Left = padding;
                dgvRes.Top = rtbInput.Bottom + padding;
                dgvRes.Width = menuStrip.Width - 2 * padding;
                dgvRes.Height = statusStrip.Top - rtbInput.Height - menuStrip.Size.Height - 3 * padding;

            }
        }

        #endregion

        #region rtbInput

        /// <summary>
        /// input comfort function
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void rtbInput_TextChanged(object sender, EventArgs e)
        {

            //only if auto completion is active
            if (autocompletionToolStripMenuItem.Checked == false)
                return;

            //auto completion only if cursor is at the end of .text
            if (rtbInput.SelectionStart != rtbInput.Text.Length)
                return;

            //############START AUTOCOMPLETION###############//
            //if operator is at left of input then add "ans"
            //only if in singleline mode
            if (!multilineToolStripMenuItem.Checked)
            {
                CalcObjectInfo.objectInfo firstObject = CalcObjectInfo.Is_Operator_At_Pos(rtbInput.Text, 0);
                if (firstObject.found && rtbInput.Text == firstObject.objectStr)
                {
                    rtbInput.Text = rtbInput.Text.Insert(0, "ans");
                    rtbInput.SelectionStart = rtbInput.Text.Length;
                }
            }
            //############END AUTOCOMPLETION###############//
        }

        #endregion

        #region Menu_Items_Insert_Etc

        #region Angle_Mode

        //switch to degree mode
        private void degreeToolStripMenuItem_Click(object sender, EventArgs e)
        {
            Set_Angle_Mode_Degree();
            Add_Calc_List_Entry("Changed angle mode to degree");
        }

        //switch to rad mode
        private void radianToolStripMenuItem_Click(object sender, EventArgs e)
        {
            Set_Angle_Mode_Radiant();
            Add_Calc_List_Entry("Changed angle mode to radian");
        }

        private void Set_Angle_Mode_Radiant()
        {
            degreeToolStripMenuItem.Checked = false;
            radianToolStripMenuItem.Checked = true;
            CalcEngine.Set_Angle_Mode(CalcEngine.angleModes.Radian);
            toolStripStatusLabelAngleMode.Text = "Rad";
        }

        private void Set_Angle_Mode_Degree()
        {
            radianToolStripMenuItem.Checked = false;
            degreeToolStripMenuItem.Checked = true;
            CalcEngine.Set_Angle_Mode(CalcEngine.angleModes.Degree);
            toolStripStatusLabelAngleMode.Text = "Deg";
        }

        #endregion

        private void Reset_Status_Label()
        {
            toolStripStatusLabelState.Text = "Ready";
        }

        private void quitToolStripMenuItem_Click(object sender, EventArgs e)
        {   Application.Exit(); }

        private void piToolStripMenuItem_Click(object sender, EventArgs e)
        {
            Input_Str_Insertion(CalcGlobals.PI.ToString(CalcEngine.Get_Decimal_Sign().NumberFormat));
        }

        private void eToolStripMenuItem_Click(object sender, EventArgs e)
        {
            Input_Str_Insertion(CalcGlobals.Euler.ToString(CalcEngine.Get_Decimal_Sign().NumberFormat));
        }

        private void phiToolStripMenuItem_Click(object sender, EventArgs e)
        {
            Input_Str_Insertion(CalcGlobals.Phi.ToString(CalcEngine.Get_Decimal_Sign().NumberFormat));
        }

        private void g981ToolStripMenuItem_Click(object sender, EventArgs e)
        {
            Input_Str_Insertion(CalcGlobals.GravEarth.ToString(CalcEngine.Get_Decimal_Sign().NumberFormat));
        }

        private void c3108ToolStripMenuItem_Click(object sender, EventArgs e)
        {
            Input_Str_Insertion(CalcGlobals.LightSpeed.ToString(CalcEngine.Get_Decimal_Sign().NumberFormat));
        }

        private void eFieldToolStripMenuItem_Click(object sender, EventArgs e)
        {
            Input_Str_Insertion(CalcGlobals.ElField.ToString(CalcEngine.Get_Decimal_Sign().NumberFormat));
        }

        private void eulerMascheroniToolStripMenuItem_Click(object sender, EventArgs e)
        {
            Input_Str_Insertion(CalcGlobals.EulerMascharoni.ToString(CalcEngine.Get_Decimal_Sign().NumberFormat));
        }

        private void bFieldToolStripMenuItem_Click(object sender, EventArgs e)
        {
            Input_Str_Insertion(CalcGlobals.MagField.ToString(CalcEngine.Get_Decimal_Sign().NumberFormat));
        }

        private void elementaryChargeToolStripMenuItem_Click(object sender, EventArgs e)
        {
            Input_Str_Insertion(CalcGlobals.ElecCharge.ToString(CalcEngine.Get_Decimal_Sign().NumberFormat));
        }

        private void plusToolStripMenuItem_Click(object sender, EventArgs e)
        {
            Input_Str_Insertion("+");
        }

        private void minusToolStripMenuItem_Click(object sender, EventArgs e)
        {
            Input_Str_Insertion("-");
        }

        private void divisionToolStripMenuItem_Click(object sender, EventArgs e)
        {
            Input_Str_Insertion("/");
        }

        private void multiplicationToolStripMenuItem_Click(object sender, EventArgs e)
        {
            Input_Str_Insertion("*");
        }

        private void radicationToolStripMenuItem_Click(object sender, EventArgs e)
        {
            Input_Str_Insertion("root");
        }

        private void powerToolStripMenuItem_Click(object sender, EventArgs e)
        {
            Input_Str_Insertion("^");
        }

        private void logarithmToolStripMenuItem_Click(object sender, EventArgs e)
        {
            Input_Str_Insertion("log");
        }

        private void sinusToolStripMenuItem_Click(object sender, EventArgs e)
        {
            Input_Str_Insertion("sin");
        }

        private void cosinusToolStripMenuItem_Click(object sender, EventArgs e)
        {
            Input_Str_Insertion("cos");
        }

        private void tangentToolStripMenuItem_Click(object sender, EventArgs e)
        {
            Input_Str_Insertion("tan");
        }

        private void arcustangentToolStripMenuItem_Click(object sender, EventArgs e)
        {
            Input_Str_Insertion("atan");
        }

        private void modulusModToolStripMenuItem_Click(object sender, EventArgs e)
        {
            Input_Str_Insertion("mod");
        }

        private void valueAssignmentVarNameValueToolStripMenuItem_Click(object sender, EventArgs e)
        {
            Input_Str_Insertion("=");
        }

        private void sincSincxToolStripMenuItem_Click(object sender, EventArgs e)
        {
            Input_Str_Insertion("=");
        }

        private void deltaxToolStripMenuItem_Click(object sender, EventArgs e)
        {
            Input_Str_Insertion("=");
        }

        private void sigmaToolStripMenuItem_Click(object sender, EventArgs e)
        {
            Input_Str_Insertion("=");
        }

        private void factoryFactxToolStripMenuItem_Click(object sender, EventArgs e)
        {
            Input_Str_Insertion("=");
        }

        private void logarithmusnaturalisLnxToolStripMenuItem_Click(object sender, EventArgs e)
        {
            Input_Str_Insertion("=");
        }

        private void logarithmusdecimalisLgxToolStripMenuItem_Click(object sender, EventArgs e)
        {
            Input_Str_Insertion("=");
        }

        private void logarithmusdualisLdxToolStripMenuItem_Click(object sender, EventArgs e)
        {
            Input_Str_Insertion("=");
        }

        private void deleteVariableToolStripMenuItem_Click(object sender, EventArgs e)
        { CalcVariables.Del_Vars(); }

        private void printAllVariablesToolStripMenuItem_Click(object sender, EventArgs e)
        { MessageBox.Show(CalcVariables.Print_Vars(), "Print variables result"); }

        private void clearListToolStripMenuItem_Click(object sender, EventArgs e)
        { clearList(null); }

        private void absoluteXAbsYToolStripMenuItem_Click(object sender, EventArgs e)
        {
            Input_Str_Insertion("abs");
        }

        private void functionDefinitionFxYToolStripMenuItem_Click(object sender, EventArgs e)
        {   Input_Str_Insertion("( ) = ");  }

        private void deleteallFunctionsToolStripMenuItem_Click(object sender, EventArgs e)
        { CalcFunctions.Del_Functions(); }

        private void printAllFunctionsToolStripMenuItem_Click(object sender, EventArgs e)
        { MessageBox.Show(CalcFunctions.Print_User_Functions(), "Print function result"); }

        private void drawFunctionToolStripMenuItem_Click(object sender, EventArgs e)
        { Open_DrawForm(); }

        private void Open_DrawForm()
        {
            if (frmDrawFct == null)
            {
                frmDrawFct = new frmDrawFunction();
                frmDrawFct.Show();
            }
            else if (!Is_Form_Open(frmDrawFct.Name))
            {
                frmDrawFct = new frmDrawFunction();
                frmDrawFct.Show();
            }
            else
            {
                frmDrawFct.Focus();
            }
        }

        private void answerAnsToolStripMenuItem_Click(object sender, EventArgs e)
        {
            Input_Str_Insertion("ans");
        }

        private void arcussinusXAsinYToolStripMenuItem_Click(object sender, EventArgs e)
        {
            Input_Str_Insertion("asin");
        }

        private void arcuscosinusXAcosYToolStripMenuItem_Click(object sender, EventArgs e)
        {
            Input_Str_Insertion("acos");
        }

        private void hyperbelsinusXHsinYToolStripMenuItem_Click(object sender, EventArgs e)
        {
            Input_Str_Insertion("sinh");
        }

        private void hyperbelcosinusXHcosYToolStripMenuItem_Click(object sender, EventArgs e)
        {
            Input_Str_Insertion("cosh");
        }

        private void hyperbeltangentXHtanYToolStripMenuItem_Click(object sender, EventArgs e)
        {
            Input_Str_Insertion("tanh");
        }

        private void roundXRndYToolStripMenuItem_Click(object sender, EventArgs e)
        {
            Input_Str_Insertion("rnd");
        }

        private void randomNumberRandToolStripMenuItem_Click(object sender, EventArgs e)
        {
            Input_Str_Insertion("rand");
        }

        private void sqarerootSqrtxToolStripMenuItem_Click(object sender, EventArgs e)
        {
            Input_Str_Insertion("sqrt");
        }

        private void rndToolStripMenuItem_Click(object sender, EventArgs e)
        {
            Input_Str_Insertion("rnd");
        }

        private void logarithmXLogYToolStripMenuItem_Click(object sender, EventArgs e)
        {
            Input_Str_Insertion("log");
        }

        private void bitshiftLeftXYToolStripMenuItem_Click(object sender, EventArgs e)
        {
            Input_Str_Insertion("<<");
        }

        private void bitshiftRightXYToolStripMenuItem_Click(object sender, EventArgs e)
        {
            Input_Str_Insertion(">>");
        }

        private void rotationLeftXRrotYToolStripMenuItem_Click(object sender, EventArgs e)
        {
            Input_Str_Insertion("lrot");
        }

        private void rotationRightXRrotYToolStripMenuItem_Click(object sender, EventArgs e)
        {
            Input_Str_Insertion("rrot");
        }

        private void lesserXYToolStripMenuItem_Click(object sender, EventArgs e)
        {
            Input_Str_Insertion("<");
        }

        private void lesserOrEqualXYToolStripMenuItem_Click(object sender, EventArgs e)
        {
            Input_Str_Insertion("<=");
        }

        private void greaterXYToolStripMenuItem_Click(object sender, EventArgs e)
        {
            Input_Str_Insertion(">");
        }

        private void greaterOrEqualXYToolStripMenuItem_Click(object sender, EventArgs e)
        {
            Input_Str_Insertion(">=");
        }

        private void equalXYToolStripMenuItem_Click(object sender, EventArgs e)
        {
            Input_Str_Insertion("==");
        }

        private void notEqualXYToolStripMenuItem_Click(object sender, EventArgs e)
        {
            Input_Str_Insertion("!=");
        }

        private void bitwiseAndXYToolStripMenuItem_Click(object sender, EventArgs e)
        {
            Input_Str_Insertion("&");
        }

        private void bitwiseOrXYToolStripMenuItem_Click(object sender, EventArgs e)
        {
            Input_Str_Insertion("|");
        }

        private void bitwiseXOrXXorYToolStripMenuItem_Click(object sender, EventArgs e)
        {
            Input_Str_Insertion("xor");
        }

        private void logicAndXYToolStripMenuItem_Click(object sender, EventArgs e)
        {
            Input_Str_Insertion("&&");
        }

        private void logicOrXYToolStripMenuItem_Click(object sender, EventArgs e)
        {
            Input_Str_Insertion("||");
        }

        private void notxToolStripMenuItem_Click(object sender, EventArgs e)
        {
            Input_Str_Insertion("!");
        }

        private void negatexToolStripMenuItem_Click(object sender, EventArgs e)
        {
            Input_Str_Insertion("~");
        }

        //add menu items
        private void deleteVariableToolStripMenuItem1_DropDownOpening(object sender, EventArgs e)
        {
            int numDelVars = delVarMenuItems.Count();
            int numVars = CalcVariables.Count();

            //clear items
            //here because the dropdownclosed-event would be fired before the item_click-event
            for (int i = 0; i < numDelVars; ++i)
            {
                if (delVarMenuItems == null)
                    break;
                if (delVarMenuItems[i] == null)
                    break;
                delVarMenuItems[i].Click -= new EventHandler(this.delVarMenuItems_Click);
                deleteVariableToolStripMenuItem1.DropDownItems.Remove(delVarMenuItems[i]);
            }
            delVarMenuItems.Clear();

            //add items
            for (int i = 0; i < numVars; ++i)
            {
                string printVar = CalcVariables.Print_Var(i);
                if (printVar == "")
                    break;
                ToolStripMenuItem newDelVarMenuItem = new ToolStripMenuItem(printVar);
                newDelVarMenuItem.Click += new EventHandler(this.delVarMenuItems_Click);
                delVarMenuItems.Add(newDelVarMenuItem);
                deleteVariableToolStripMenuItem1.DropDownItems.Add(newDelVarMenuItem);
            }
        }
        
        //delete var shown on item
        private void delVarMenuItems_Click(object sender, EventArgs e)
        {
            int numVars = delVarMenuItems.Count();

            for (int i = 0; i < numVars; ++i)
            {
                if (delVarMenuItems[i] == null)
                    break;
                if (sender.Equals(delVarMenuItems[i]))
                {
                    CalcVariables.Del_Var(i);
                }
            }
        }

        //add menu items
        private void deleteFunctionToolStripMenuItem_DropDownOpening(object sender, EventArgs e)
        {
            int numDelFcts = delFctMenuItems.Count();
            int numFcts = CalcFunctions.Count();

            //clear items
            //here because the dropdownclosed-event would be fired before the item_click-event
            for (int i = 0; i < numDelFcts; ++i)
            {
                if (delFctMenuItems == null)
                    break;
                if (delFctMenuItems[i] == null)
                    break;
                delFctMenuItems[i].Click -= new EventHandler(this.delFctMenuItems_Click);
                delFctToolStripMenuItem.DropDownItems.Remove(delFctMenuItems[i]);
            }
            delFctMenuItems.Clear();

            //add items
            for (int i = 0; i < numFcts; ++i)
            {
                //skip standard functions
                if (!CalcFunctions.Is_Std_Fct(i))
                {
                    string printFct = CalcFunctions.Print_Function(i);
                    if (printFct == "")
                        break;
                    ToolStripMenuItem newDelFctMenuItem = new ToolStripMenuItem(printFct);
                    newDelFctMenuItem.Click += new EventHandler(this.delFctMenuItems_Click);
                    //use tag as container for index of containing function to delete
                    //cannot use index of delfctmenuitems because of missing standard functions
                    newDelFctMenuItem.Tag = i;      
                    delFctMenuItems.Add(newDelFctMenuItem);
                    delFctToolStripMenuItem.DropDownItems.Add(newDelFctMenuItem);
                }
            }
        }

        //delete var shown on item
        private void delFctMenuItems_Click(object sender, EventArgs e)
        {
            int numFcts = delFctMenuItems.Count;

            for (int i = 0; i < numFcts; ++i)
            {
                if (delFctMenuItems[i] == null)
                    break;
                if (sender.Equals(delFctMenuItems[i]))
                {
                    CalcFunctions.Del_Function((int)delFctMenuItems[i].Tag);
                }
            }
        }

        //switch auto completion state
        private void autocompletionToolStripMenuItem_Click(object sender, EventArgs e)
        {
            if (autocompletionToolStripMenuItem.Checked)
                autocompletionToolStripMenuItem.Checked = false;
            else
                autocompletionToolStripMenuItem.Checked = true;
        }

        /// <summary>
        /// check if round dezimals is in valid area
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void toolStripTextBox1_TextChanged(object sender, EventArgs e)
        {
            Verify_Round_Decimals();
        }

        private void writeLogToolStripMenuItem_Click(object sender, EventArgs e)
        {
            if (writeLogToolStripMenuItem.Checked)
                writeLogToolStripMenuItem.Checked = false;
            else
                writeLogToolStripMenuItem.Checked = true;

            CalcEngine.Set_Write_Log(writeLogToolStripMenuItem.Checked);
        }

        private void setOutputFormat(CalcEngine.OutputFormat oF)
        {
            if (oF == CalcEngine.OutputFormat.Binary)
                binaryToolStripMenuItem.Checked = true;
            else
                binaryToolStripMenuItem.Checked = false;

            if (oF == CalcEngine.OutputFormat.Octal)
                octalToolStripMenuItem.Checked = true;
            else
                octalToolStripMenuItem.Checked = false;

            if (oF == CalcEngine.OutputFormat.Decimal)
                decimalToolStripMenuItem.Checked = true;
            else
                decimalToolStripMenuItem.Checked = false;

            if (oF == CalcEngine.OutputFormat.Exponential)
                exponentialViewToolStripMenuItem.Checked = true;
            else
                exponentialViewToolStripMenuItem.Checked = false;

            if (oF == CalcEngine.OutputFormat.Hexadecimal)
                hexadecimalToolStripMenuItem.Checked = true;
            else
                hexadecimalToolStripMenuItem.Checked = false;
        }

        private void exponentialViewToolStripMenuItem_Click(object sender, EventArgs e)
        {
            setOutputFormat(CalcEngine.OutputFormat.Exponential);
        }

        private void decimalToolStripMenuItem_Click(object sender, EventArgs e)
        {
            setOutputFormat(CalcEngine.OutputFormat.Decimal);
        }

        private void binaryToolStripMenuItem_Click(object sender, EventArgs e)
        {
            setOutputFormat(CalcEngine.OutputFormat.Binary);
        }

        private void octalToolStripMenuItem_Click(object sender, EventArgs e)
        {
            setOutputFormat(CalcEngine.OutputFormat.Octal);
        }

        private void hexadecimalToolStripMenuItem_Click(object sender, EventArgs e)
        {
            setOutputFormat(CalcEngine.OutputFormat.Hexadecimal);
        }

        private void loadFileToolStripMenuItem_Click(object sender, EventArgs e)
        {
            OpenFileDialog oFDia = new OpenFileDialog();

            oFDia.Filter = "QCalc files|*.qcalc|Text files|*.txt|All files|*.*";
            oFDia.FilterIndex = 1;

            if (oFDia.ShowDialog() != DialogResult.OK)
                return;

            Load_File(oFDia.FileName);
        }

        private void MainForm_DragEnter(object sender, DragEventArgs e)
        {
            if (e.Data.GetDataPresent(DataFormats.FileDrop) || e.Data.GetDataPresent(DataFormats.Text)) 
                e.Effect = DragDropEffects.Copy;
        }

        private void Execute_Line(string line)
        {
            rtbInput.Text = line;
            Start_Calculation();
        }

        #endregion

        private void aboutToolStripMenuItem_Click(object sender, EventArgs e)
        {
            frmAbout = new About();
            frmAbout.Show();
        }

        private void readMeFileToolStripMenuItem_Click(object sender, EventArgs e)
        {
            try
            {
                System.Diagnostics.Process.Start(Application.StartupPath + @"/QuickCalculator ReadMeOrNot.html");
            }
            catch (System.ComponentModel.Win32Exception ex)
            {
                MessageBox.Show("Error while openening readme file: " + ex.Message);
            }
            catch (System.IO.FileNotFoundException ex)
            {
                MessageBox.Show("File not found error: " + ex.Message);
            }
            catch (Exception ex)
            {
                MessageBox.Show("Unknown error: " + ex.Message);
            }
        }

        #region DecimalSign
        private void commaToolStripMenuItem_Click(object sender, EventArgs e)
        {
            Set_DecSign_Comma();
        }

        private void pointToolStripMenuItem_Click(object sender, EventArgs e)
        {
            Set_DecSign_Point();
        }

        private void Set_DecSign_Point()
        {
            CalcEngine.Set_Decimal_Sign(CalcEngine.DecimalSign.point);
            Actualize_DecSign();
        }

        private void Set_DecSign_Comma()
        {
            CalcEngine.Set_Decimal_Sign(CalcEngine.DecimalSign.comma);
            Actualize_DecSign();
        }
        #endregion

    }
}
